# -*- coding: utf-8 -*-
"""
Created on Thu Apr 16 15:30:38 2015

@author: noore
"""
import numpy as np


class NonSteadyStateSolutionError(Exception):
    pass


class ThermodynamicallyInfeasibleError(Exception):
    def __init__(self, lnC=None):
        if lnC is None:
            Exception.__init__(
                self, "this reaction system has no feasible solutions"
            )
        else:
            Exception.__init__(
                self,
                "C = %s : is thermodynamically infeasible" % str(np.exp(lnC)),
            )

    pass



"""analyze pathways using thermodynamic models."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import csv
import logging
import warnings
from typing import Callable, Dict, Iterable, List, TextIO, Tuple

import numpy as np
import pandas as pd
from equilibrator_api import (
    Q_,
    Compound,
    R,
    ccache,
    default_I,
    default_pH,
    default_pMg,
    default_T,
    search_reaction,
)
from equilibrator_api.component_contribution import ComponentContribution
from equilibrator_api.phased_reaction import PhasedReaction
from equilibrator_cache.reaction import (
    create_stoichiometric_matrix_from_reactions,
)
from sbtab import SBtab

from .bounds import Bounds
from .thermo_models import PathwayMDFData, PathwayThermoModel


class Pathway(object):
    """A pathway parsed from user input.

    Designed for checking input prior to converting to a stoichiometric model.
    """

    def __init__(
        self,
        reactions: List[PhasedReaction],
        fluxes: np.ndarray,
        standard_dg_primes: np.ndarray = None,
        dg_sigma: np.ndarray = None,
        bounds: Bounds = None,
        config_dict: Dict[str, str] = None,
    ) -> object:
        """Initialize.

        :param reactions: a list of gibbs.reaction.Reaction objects.
        :param fluxes: numpy.array of relative fluxes in same order as
        reactions.
        :param standard_dg_primes: reaction energies (in kJ/mol)
        :param dg_sigma: square root of the uncertainty covariance matrix
        (in kJ/mol)
        :param bounds: bounds on metabolite concentrations. Uses default
        bounds if None provided.
        """
        self.reactions = reactions
        Nr = len(reactions)

        if bounds is None:
            self._bounds = Bounds.GetDefaultBounds().Copy()
        else:
            self._bounds = bounds.Copy()

        self.p_h = default_pH
        self.p_mg = default_pMg
        self.ionic_strength = default_I
        self.temperature = default_T
        self.config_dict = config_dict or dict()
        self.configure()

        # make sure the bounds on Water are (1, 1)
        self._bounds.SetBounds(ccache.water, Q_(1.0, "M"), Q_(1.0, "M"))

        self.S = create_stoichiometric_matrix_from_reactions(
            reactions, ccache.is_proton, ccache.is_water, ccache.water
        )

        assert fluxes.unitless or fluxes.check("[concentration]/[time]")
        self.fluxes = fluxes.flatten()
        assert self.fluxes.shape == (Nr,)

        self.comp_contrib = None
        if standard_dg_primes is None:
            assert dg_sigma is None, (
                "If standard_dg_primes are not "
                "provided, dg_sigma must also be None"
            )
            self.update_standard_dgs()
        else:
            assert standard_dg_primes.check("[energy]/[substance]")
            assert standard_dg_primes.shape == (Nr,)
            # dGr should be orthogonal to nullspace of S
            # If not, dGr is not contained in image(S) and then there
            # is no consistent set of dGfs that generates dGr and the
            # first law of thermo is violated by the model.
            S_T = self.S.T.values
            S_inv = np.linalg.pinv(S_T)
            null_proj = np.eye(self.S.shape[1]) - S_T @ S_inv
            projected = null_proj @ standard_dg_primes.T
            assert (projected < Q_("1e-8 kJ/mol")).all(), (
                "Supplied reaction standard deltaG values are inconsistent "
                "with the stoichiometric matrix."
            )

            self.standard_dg_primes = standard_dg_primes

            if dg_sigma is None:
                self.dg_sigma = None
            else:
                assert dg_sigma.shape == (Nr, Nr)
                self.dg_sigma = dg_sigma

        # Set the compound names by to their default strings, these names can
        # be changed later using the same function (set_compound_names)
        self.compound_names = None
        self.set_compound_names(str)

    def configure(self) -> None:
        if "p_h" in self.config_dict:
            self.p_h = Q_(self.config_dict["p_h"])
            assert self.p_h.check(""), "pH must be given without units"

        if "p_mg" in self.config_dict:
            self.p_mg = Q_(self.config_dict["p_mg"])
            assert self.p_mg.check(""), "pMg must be given without units"

        if "ionic_strength" in self.config_dict:
            self.ionic_strength = Q_(self.config_dict["ionic_strength"])
            assert self.ionic_strength.check(
                "[concentration]"
            ), "ionic_strength must be given in concentration units"

        if "temperature" in self.config_dict:
            self.temperature = Q_(self.config_dict["temperature"])
            assert self.temperature.check(
                "[temperature]"
            ), "temperature must be given in Kalvin"

        # this is the factor by which the STDEV matrix is multiplied to create
        # the confidence margins for the Gibbs energies of reaction
        self.stdev_factor = float(self.config_dict.get("stdev_factor", "0"))

    def update_standard_dgs(self,) -> None:
        """Set the aqueous conditions and recalculate the standard dG' values.
        """
        if self.comp_contrib is None:
            self.comp_contrib = ComponentContribution(
                p_h=self.p_h,
                p_mg=self.p_mg,
                ionic_strength=self.ionic_strength,
                temperature=self.temperature,
            )
        else:
            self.comp_contrib.p_h = self.p_h
            self.comp_contrib.p_mg = self.p_mg
            self.comp_contrib.ionic_strength = self.ionic_strength
            self.comp_contrib.temperature = self.temperature

        (
            self.standard_dg_primes,
            self.dg_sigma,
        ) = self.comp_contrib.standard_dg_prime_multi(self.reactions)

    def set_compound_names(self, mapping: Callable[[Compound], str]) -> None:
        """Use alternative compound names for outputs such as plots.

        :param mapping: a dictionary mapping compounds to their names in the
        model
        """
        self.compound_names = list(map(mapping, self.S.index))

    @property
    def bounds(self) -> Tuple[Iterable[float], Iterable[float]]:
        """Get the concentration bounds.

        The order of compounds is according to the stoichiometric matrix index.
        :return: tuple of (lower bounds, upper bounds)
        """
        return self._bounds.GetBounds(self.S.index)

    @property
    def ln_conc_lb(self) -> np.array:
        """Get the log lower bounds on the concentrations.

        The order of compounds is according to the stoichiometric matrix index.
        :return: a NumPy array of the log lower bounds
        """
        return np.array(
            list(map(float, self._bounds.GetLnLowerBounds(self.S.index)))
        )

    @property
    def ln_conc_ub(self) -> np.array:
        """Get the log upper bounds on the concentrations.

        The order of compounds is according to the stoichiometric matrix index.
        :return: a NumPy array of the log upper bounds
        """
        return np.array(
            list(map(float, self._bounds.GetLnUpperBounds(self.S.index)))
        )

    @staticmethod
    def get_compounds(reactions: Iterable[PhasedReaction]) -> List[Compound]:
        """Get a unique list of all compounds in all reactions.

        :param reactions: an iterator of reactions
        :return: a list of unique compounds
        """
        compounds = set()
        for r in reactions:
            compounds.update(r.keys())

        return sorted(compounds)

    def calc_mdf(self) -> PathwayMDFData:
        """Calculate the Max-min Driving Force.

        :param stdev_factor: the factor by which to multiply the uncertainties
        :return: a PathwayMDFData object with the results
        """
        return PathwayThermoModel(self, self.stdev_factor).FindMDF()

    @property
    def reaction_ids(self) -> Iterable[str]:
        """Iterate through all the reaction IDs.

        :return: the reaction IDs
        """
        return map(lambda rxn: rxn.rid, self.reactions)

    @property
    def reaction_formulas(self) -> Iterable[str]:
        """Iterate through all the reaction formulas.

        :return: the reaction formulas
        """
        return map(str, self.reactions)

    @property
    def net_reaction(self) -> PhasedReaction:
        """Calculate the sum of all the reactions in the pathway.

        :return: the net reaction
        """
        net_rxn_stoich = self.S @ self.fluxes.magnitude
        net_rxn_stoich = net_rxn_stoich[net_rxn_stoich != 0]
        sparse = net_rxn_stoich.to_dict()
        return PhasedReaction(sparse)

    @classmethod
    def from_csv(
        cls,
        f: TextIO,
        bounds: Bounds = None,
        p_h: float = default_pH,
        p_mg: float = default_pMg,
        ionic_strength: float = default_I,
        temperature: float = default_T,
    ) -> object:
        """Parse a CSV file and return a Pathway object.

        Caller responsible for closing f.

        :param f: file-like object containing CSV data describing the pathway
        :param bounds: a Bounds object
        :param p_h:
        :param p_mg:
        :param ionic_strength:
        :param: temperature:
        :return: a Pathway object
        """
        reactions = []
        fluxes = []

        for row in csv.DictReader(f):
            rxn_formula = row.get("ReactionFormula")

            flux = float(row.get("Flux", 0.0))
            logging.debug("formula = %f x (%s)", flux, rxn_formula)

            rxn = search_reaction(rxn_formula)
            rxn.is_balanced(ignore_atoms=["H"])

            reactions.append(rxn)
            fluxes.append(flux)

        v = np.array(fluxes)

        config_dict = {
            "p_h": str(p_h),
            "p_mg": str(p_mg),
            "ionic_strength": str(ionic_strength),
            "temperature": str(temperature),
        }
        pp = Pathway(reactions, v, bounds=bounds, config_dict=config_dict)
        return pp

    @staticmethod
    def from_sbtab(sbtabdoc: SBtab.SBtabDocument) -> object:
        """Parse and SBtabDocument and return a Pathway.

        :param sbtabdoc: an SBtabDocument defining the pathway
        :return: a Pathway object
        """

        # Read the configuration table if it exists
        config_sbtab = sbtabdoc.get_sbtab_by_id("Configuration")
        if config_sbtab:
            config_df = config_sbtab.to_data_frame()
            assert (
                "Option" in config_df.columns
            ), "Configuration table must have an Option column"
            assert (
                "Value" in config_df.columns
            ), "Configuration table must have an Value column"
            config_dict = config_df.set_index("Option").Value.to_dict()
        else:
            config_dict = dict()

        table_ids = ["Compound", "ConcentrationConstraint", "Reaction", "Flux"]
        dfs = []

        for table_id in table_ids:
            sbtab = sbtabdoc.get_sbtab_by_id(table_id)
            if sbtab is None:
                tables = ", ".join(map(lambda s: s.table_id, sbtabdoc.sbtabs))
                raise ValueError(
                    f"The SBtabDocument must have a table "
                    f"with the following ID: {table_id}, "
                    f"however, only these tables were "
                    f"found: {tables}"
                )
            dfs.append(sbtab.to_data_frame())

        compound_df, bounds_df, reaction_df, flux_df = dfs

        # Read the Compound table
        # -----------------------
        # use equilibrator-cache to build a dictionary of compound IDs to
        # Compound objects
        compound_df.set_index("ID", inplace=True)

        # legacy option, KEGG IDs could be provided without the namespace
        # when the column title  explicitly indicated that they were from KEGG
        if "Identifiers:kegg.compound" in compound_df.columns:
            compound_df["Identifiers"] = (
                "kegg:" + compound_df["Identifiers:kegg.compound"]
            )

        # Now, by default, we assume the namespaces are provided as part of the
        # reaction formulas
        if "Identifiers" not in compound_df.columns:
            raise KeyError(
                "There is no column of Identifiers in the Compound table"
            )

        compound_df["Compound"] = compound_df["Identifiers"].apply(
            ccache.get_compound
        )

        if pd.isnull(compound_df.Compound).any():
            accessions_not_found = compound_df.loc[
                pd.isnull(compound_df.Compound), "Identifiers"
            ]
            error_msg = str(accessions_not_found.to_dict())
            raise KeyError(
                f"Some compounds not found in equilibrator-cache: "
                f"{error_msg}"
            )

        # Read the ConcentrationConstraints table
        # ---------------------------------------
        # convert compound IDs in the bounds table to Compound objects
        name_to_compound = compound_df.Compound.to_dict()
        bounds_df["Compound"] = bounds_df.Compound.apply(name_to_compound.get)
        bounds_df.set_index("Compound", inplace=True)

        bounds_sbtab = sbtabdoc.get_sbtab_by_id("ConcentrationConstraint")
        try:
            bounds_unit = bounds_sbtab.get_attribute("Unit")
            lbs = bounds_df["Min"].apply(lambda x: Q_(float(x), bounds_unit))
            ubs = bounds_df["Max"].apply(lambda x: Q_(float(x), bounds_unit))
        except SBtab.SBtabError:
            # if the unit is not defined in the header, we assume it is given
            # next to each bound individually
            lbs = bounds_df["Min"].apply(Q_)
            ubs = bounds_df["Max"].apply(Q_)

        bounds = Bounds(lbs.to_dict(), ubs.to_dict())
        bounds._check_bounds()

        # Read the Reaction table
        # -----------------------
        reactions = []
        reaction_ids = []
        for row in reaction_df.itertuples():
            rxn = PhasedReaction.parse_formula(
                name_to_compound.get, row.ReactionFormula
            )
            rxn.rid = row.ID
            if not rxn.is_balanced(ignore_atoms=("H",)):
                warnings.warn(f"Reaction {rxn.rid} is not balanced")
            reactions.append(rxn)
            if row.ID in reaction_ids:
                raise KeyError(
                    f"Reaction IDs must be unique, but you have "
                    f"{row.ID} twice"
                )
            reaction_ids.append(row.ID)

        # Read the Flux table
        # ---------------------------
        flux_sbtab = sbtabdoc.get_sbtab_by_id("Flux")
        fluxes = (
            flux_df.set_index("Reaction")
            .loc[reaction_ids, "Value"]
            .apply(float)
        )

        fluxes = np.array(fluxes.values, ndmin=2, dtype=float).T

        try:
            # convert fluxes to M/s if they are in some other absolute unit
            flux_unit = flux_sbtab.get_attribute("Unit")
            fluxes *= Q_(1.0, flux_unit)
        except SBtab.SBtabError:
            # otherwise, assume these are relative fluxes
            fluxes *= Q_("dimensionless")

        thermo_sbtab = sbtabdoc.get_sbtab_by_id("Thermodynamics")
        if thermo_sbtab:
            rid2dg0 = Pathway.ReadThermodynamics(thermo_sbtab, config_dict)

            # make sure all reactions have a standard Gibbs energy
            assert set(reaction_ids).issubset(rid2dg0.keys()), (
                "Not all reactions have a equilibrium constant or gibbs energy "
                "in the 'Thermodynamics' table."
            )
            standard_dg_primes = np.array(
                [rid2dg0[rid].m_as("kJ/mol") for rid in reaction_ids]
            ) * Q_("1 kJ/mol")

            pp = Pathway(
                reactions,
                fluxes=fluxes,
                standard_dg_primes=standard_dg_primes,
                dg_sigma=None,
                bounds=bounds,
                config_dict=config_dict,
            )
        else:
            pp = Pathway(
                reactions, fluxes=fluxes, bounds=bounds, config_dict=config_dict
            )

        # override the compound names with the ones in the SBtab
        compound_to_name = dict(map(reversed, name_to_compound.items()))
        pp.set_compound_names(compound_to_name.get)
        return pp

    @staticmethod
    def ReadThermodynamics(
        thermo_sbtab: SBtab.SBtabTable, config_dict: Dict[str, str]
    ) -> Dict[str, Q_]:

        try:
            std_conc = thermo_sbtab.get_attribute("StandardConcentration")
            assert Q_(std_conc) == Q_(
                "1 M"
            ), "We only support a standard concentration of 1 M."
        except SBtab.SBtabError:
            pass

        if "temperature" in config_dict:
            temperature = Q_(config_dict["temperature"])
            assert temperature.check("[temperature]")
        else:
            temperature = default_T

        thermo_df = thermo_sbtab.to_data_frame()
        cols = ["QuantityType", "Value", "Compound", "Reaction", "Unit"]

        rid2dg0 = dict()
        for i, row in thermo_df.iterrows():
            try:
                typ, val, cid, rid, unit = [row[c] for c in cols]
                val = Q_(float(val), unit)

                if typ == "equilibrium constant":
                    assert val.check(
                        ""
                    ), f"equilibrium constants must have no units"
                    rid2dg0[rid] = -np.log(val.m_as("")) * R * temperature
                elif typ == "reaction gibbs energy":
                    assert val.check(
                        "[energy]/[substance]"
                    ), f"Gibbs energies must be in kJ/mol (or equivalent)"
                    val.ito("kJ/mol")
                    rid2dg0[rid] = val
                else:
                    raise AssertionError(
                        "unrecognized Rate Constant Type: " + typ
                    )
            except AssertionError:
                raise ValueError(
                    "Syntax error in Parameter table, row %d - %s" % (i, row)
                )

        return rid2dg0



"""thermo_models contains tools for running MDF and displaying results."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


import logging
from typing import Iterable, List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from equilibrator_api import Q_, standard_concentration, ureg
from matplotlib.lines import Line2D
from optlang.glpk_interface import Constraint, Model, Objective, Variable
from scipy.linalg import fractional_matrix_power

from .util import RT


class PathwayMDFData(object):
    """Handle MDF results.

    PathwayMDFData is a container class for MDF results, with plotting
    capabilities.
    """

    def __init__(
        self,
        thermo_model,
        B: float,
        I_dir: np.array,
        lnC: np.array,
        y: np.array,
        reaction_prices: np.array,
        compound_prices: np.array,
    ) -> object:
        """Create PathwayMDFData object.

        :param pathway: a Pathway object
        :param B: the MDF value (in units of RT)
        :param I_dir: matrix of flux directions
        :param lnC: log concentrations at MDF optimum
        :param y: uncertainty matrix coefficients at MDF optimum
        :param reaction_prices: shadow prices of reactions
        :param compound_prices: shadow prices of compound concentrations
        """
        self.mdf = B * RT
        self.bounds = thermo_model.pathway._bounds.Copy()

        concentrations = np.exp(lnC) * standard_concentration

        standard_dg_primes_over_rt = thermo_model.standard_dg_primes_over_rt

        physiological_dg_primes_over_rt = (
            standard_dg_primes_over_rt
            + PathwayMDFData.phys_dg_correction(thermo_model.pathway)
        )

        optimized_dg_primes_over_rt = (
            standard_dg_primes_over_rt
            + PathwayMDFData.dg_correction(thermo_model.pathway, concentrations)
        )

        # add the calculated error values (due to the dG'0 uncertainty)
        if thermo_model.dg_sigma_over_rt is not None:
            optimized_dg_primes_over_rt += thermo_model.dg_sigma_over_rt @ y

        # adjust dG to flux directions and convert to kJ/mol
        standard_dg_primes = (I_dir @ standard_dg_primes_over_rt) * RT
        physiological_dg_prime = (I_dir @ physiological_dg_primes_over_rt) * RT
        optimized_dg_prime = (I_dir @ optimized_dg_primes_over_rt) * RT

        # all dG values are in units of RT, so we convert them to kJ/mol
        reaction_data = zip(
            thermo_model.pathway.reaction_ids,
            thermo_model.pathway.reaction_formulas,
            thermo_model.pathway.fluxes,
            standard_dg_primes,
            physiological_dg_prime,
            optimized_dg_prime,
            reaction_prices,
        )
        self.reaction_df = pd.DataFrame(
            data=list(reaction_data),
            columns=[
                "reaction_id",
                "reaction_formula",
                "flux",
                "standard_dg_prime",
                "physiological_dg_prime",
                "optimized_dg_prime",
                "shadow_price",
            ],
        )

        compound_data = zip(
            thermo_model.pathway.compound_names, concentrations, compound_prices
        )

        self.compound_df = pd.DataFrame(
            data=list(compound_data),
            columns=["compound", "concentration", "shadow_price"],
        )
        lbs, ubs = self.bounds.GetBounds(thermo_model.pathway.compound_names)
        self.compound_df["lower_bound"] = list(lbs)
        self.compound_df["upper_bound"] = list(ubs)

    @staticmethod
    def phys_dg_correction(pathway: object) -> np.array:
        """Add the effect of reactant physiological concentrations on the dG'.

        :param pathway: Pathway object
        :return: the reaction energies (in units of RT)
        """
        dg_adj = np.array(
            [float(r.physiological_dg_correction()) for r in pathway.reactions]
        )
        return dg_adj

    @staticmethod
    @ureg.check(None, "[concentration]")
    def dg_correction(pathway: object, concentrations: np.array) -> np.array:
        """Add the effect of reactant concentrations on the dG'.

        :param pathway: Pathway object
        :param concentrations: a NumPy array of concentrations
        :return: the reaction energies (in units of RT)
        """
        log_conc = np.log(concentrations / standard_concentration)

        if np.isnan(pathway.standard_dg_primes).any():
            dg_adj = np.zeros(pathway.S.shape[1])
            for r in range(pathway.S.shape[1]):
                reactants = list(pathway.S[:, r].nonzero()[0].flat)
                dg_adj[r] = log_conc[reactants] @ pathway.S[reactants, r]
        else:
            dg_adj = pathway.S.T.values @ log_conc

        return dg_adj

    @property
    def compound_plot(self) -> plt.Figure:
        """Plot compound concentrations.

        :return: matplotlib Figure
        """
        ureg.setup_matplotlib(True)

        data_df = self.compound_df.copy()

        data_df["y"] = np.arange(0, data_df.shape[0])
        data_df["color"] = "green"

        # a sub-DataFrame of only the metabolites with nonzero shadow prices
        data_df.loc[data_df.shadow_price != 0, "color"] = "red"

        # a sub-DataFrame of only the metabolites with fixed concentrations
        data_df.loc[
            data_df.lower_bound == data_df.upper_bound, "color"
        ] = "grey"

        compound_fig, ax = plt.subplots(1, 1, figsize=(9, 6))
        ax.xaxis.set_units(ureg.molar)
        ax.axvspan(Q_("1 nM"), self.bounds.default_lb, color="y", alpha=0.5)
        ax.axvspan(self.bounds.default_ub, Q_("10M"), color="y", alpha=0.5)

        ax.scatter(
            x=data_df.concentration.tolist(),
            y=data_df.y.tolist(),
            c=data_df.color.tolist(),
        )

        ax.set_ylabel("")
        ax.set_yticks(data_df.y)
        ax.set_yticklabels(data_df["compound"], fontsize=9)
        ax.set_xlabel("Concentration (M)")
        ax.set_xscale("log")
        ax.set_ylim(-1.5, data_df.shape[0] + 0.5)
        compound_fig.tight_layout()

        # Shrink current axis by 20%
        box = ax.get_position()
        ax.set_position([box.x0, box.y0, box.width * 0.7, box.height])
        # Put a legend to the right of the current axis
        custom_lines = [
            Line2D([0], [0], color="green", lw=4),
            Line2D([0], [0], color="red", lw=4),
            Line2D([0], [0], color="grey", lw=4),
        ]
        ax.legend(
            custom_lines,
            [r"shadow price $=$ 0", r"shadow price $\ne$ 0", r"fixed value"],
            loc="center left",
            bbox_to_anchor=(1, 0.5),
            fontsize=9,
        )

        return compound_fig

    @property
    def reaction_plot(self) -> plt.Figure:
        """Plot cumulative delta-G profiles.

        :return: matplotlib Figure
        """
        ureg.setup_matplotlib(True)

        cumulative_dgms = [Q_("0 kJ/mol")] + np.cumsum(
            self.reaction_df.physiological_dg_prime
        ).tolist()
        cumulative_dgs = [Q_("0 kJ/mol")] + np.cumsum(
            self.reaction_df.optimized_dg_prime
        ).tolist()

        # This is an ugly hack. For some reason, Pandas messes up the numpy
        # arrays with pint units, and they need to be rebuilt from scratch to
        # get the units right.
        cumulative_dgms = np.array(
            [float(x / Q_("kJ/mol")) for x in cumulative_dgms]
        ) * Q_("kJ/mol")
        cumulative_dgs = np.array(
            [float(x / Q_("kJ/mol")) for x in cumulative_dgs]
        ) * Q_("kJ/mol")

        xticks = 0.5 + np.arange(self.reaction_df.shape[0])
        xticklabels = self.reaction_df.reaction_id.tolist()

        reaction_fig, ax = plt.subplots(1, 1, figsize=(9, 6))
        ax.xaxis.set_units(Q_("kJ/mol"))
        ax.plot(
            cumulative_dgms,
            label="Physiological concentrations (1 mM)",
            color="blue",
        )
        ax.plot(
            cumulative_dgs, label="MDF-optimized concentrations", color="green"
        )
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, rotation=45, ha="right")
        ax.grid(color="grey", linestyle="--", linewidth=1, alpha=0.2)

        ax.set_xlabel("Reaction Step")
        ax.set_ylabel(r"Cumulative $\Delta_r G^\prime$ (kJ/mol)")
        ax.legend(loc="best")
        ax.set_title(f"MDF = {self.mdf:.2f}")

        reaction_fig.tight_layout()
        return reaction_fig


class PathwayThermoModel(object):
    """Container for doing pathway-level thermodynamic analysis."""

    def __init__(self, pathway, stdev_factor: float = 1.0) -> object:
        """Create a model for running MDF analysis.

        :param pathway: a Pathway object defining the reactions and bounds
        :param stdev_factor: the factor with which we multiply the
        sqrt uncertainty matrix of the dG'0 estimates
        """
        self.pathway = pathway
        self.stdev_factor = stdev_factor
        self.Nc, self.Nr = pathway.S.shape

        # Make sure dG0_r' is the right size
        assert pathway.standard_dg_primes.shape == (
            self.Nr,
        ), "standard dG required for all reactions"

        self.standard_dg_primes_over_rt = (
            pathway.standard_dg_primes / RT
        ).m_as("")

        if pathway.dg_sigma is not None:
            assert pathway.dg_sigma.shape == (
                self.Nr,
                self.Nr,
            ), "uncertainty in dG required for all reactions"

            self.dg_sigma_over_rt = fractional_matrix_power(
                (pathway.dg_sigma / RT ** 2).m_as(""), 0.5
            )
        else:
            self.dg_sigma_over_rt = None

        self.I_dir = np.diag(np.sign(self.pathway.fluxes.magnitude).flat)
        self.Nr_active = np.trace(self.I_dir != 0)

        # Currently unused bounds on reaction dGs.
        self.r_bounds = None

    def _MakeLnConcentratonBounds(
        self,
    ) -> Tuple[Iterable[float], Iterable[float]]:
        """Make bounds on logarithmic concentrations.

        :return: A two-tuple (lower bounds, upper bounds).
        """
        return self.pathway.bounds.GetLnBounds(self.pathway.S.index)

    def _MakeDrivingForceConstraints(
        self,
    ) -> Tuple[np.array, np.array, np.array]:
        """Generate parameters for LP.

        Generates the A matrix and b & c vectors that can be used in a
        standard form linear problem:
                max          c'x
                subject to   Ax <= b

        x is the vector of (y | log-conc | B)
        where y dG'0 are the reaction Gibbs energy variables, log-conc
        are the natural log of the concentrations of metabolites, and
        B is the max-min driving force variable which is being maximized
        by the LP
        :return: (A, b, c) - the parameters of the LP
        """
        inds = np.nonzero(np.diag(self.I_dir))[0].tolist()

        if self.dg_sigma_over_rt is None:
            A11 = np.zeros((len(inds), self.Nr))
        else:
            A11 = self.I_dir[inds] @ self.dg_sigma_over_rt
        A12 = self.I_dir[inds] @ self.pathway.S.T
        A13 = np.ones((len(inds), 1))

        # covariance var ub and lb
        A21 = np.eye(self.Nr)
        A22 = np.zeros((self.Nr, self.Nc))
        A23 = np.zeros((self.Nr, 1))

        # log conc ub and lb
        A31 = np.zeros((self.Nc, self.Nr))
        A32 = np.eye(self.Nc)
        A33 = np.zeros((self.Nc, 1))

        # upper bound values
        b1 = -self.I_dir[inds] @ self.standard_dg_primes_over_rt
        b2 = np.ones(self.Nr) * self.stdev_factor

        A = np.vstack(
            [
                np.hstack([A11, A12, A13]),  # driving force
                np.hstack([A21, A22, A23]),  # covariance var ub
                np.hstack([-A21, A22, A23]),  # covariance var lb
                np.hstack([A31, A32, A33]),  # log conc ub
                np.hstack([A31, -A32, A33]),
            ]
        )  # log conc lb

        b = np.hstack(
            [b1, b2, b2, self.pathway.ln_conc_ub, -self.pathway.ln_conc_lb]
        )

        c = np.zeros(A.shape[1])
        c[-1] = 1.0

        # change the constraints such that reaction that have an explicit
        # r_bound will not be constrained by B, but will be constrained by
        # their specific bounds. Note that we need to divide the bound
        # by R*T since the variables in the LP are not in kJ/mol but in units
        # of R*T.
        if self.r_bounds:
            for i, r_ub in enumerate(self.r_bounds):
                if r_ub is not None:
                    A[i, -1] = 0.0
                    b[i] += float(r_ub)

        return A, b, c

    def _GetPrimalVariablesAndConstants(
        self,
    ) -> Tuple[np.array, np.array, np.array, List[Variable], List[Variable]]:

        # Create the driving force variable and add the relevant constraints
        A, b, c = self._MakeDrivingForceConstraints()

        # the dG'0 covariance eigenvariables
        y = [Variable("y%d" % i) for i in range(self.Nr)]

        # ln-concentration variables
        lnC = [Variable("l%d" % i) for i in range(self.Nc)]

        return A, b, c, y, lnC

    def _GetDualVariablesAndConstants(
        self,
    ) -> Tuple[
        np.array,
        np.array,
        np.array,
        List[Variable],
        List[Variable],
        List[Variable],
        List[Variable],
    ]:
        # Create the driving force variable and add the relevant constraints
        A, b, c = self._MakeDrivingForceConstraints()

        w = [Variable("w%d" % i, lb=0) for i in range(self.Nr_active)]
        g = [Variable("g%d" % i, lb=0) for i in range(2 * self.Nr)]
        z = [Variable("z%d" % i, lb=0) for i in range(self.Nc)]
        u = [Variable("u%d" % i, lb=0) for i in range(self.Nc)]

        return A, b, c, w, g, z, u

    def _MakeMDFProblem(
        self,
    ) -> Tuple[Model, List[Variable], List[Variable], Variable]:
        """Create primal LP problem for Min-max Thermodynamic Driving Force.

        Does not set the objective function... leaves that to the caller.

        :return: the linear problem object, and the three types of variables
        as arrays.
        """
        A, b, c, y, lnC = self._GetPrimalVariablesAndConstants()

        B = Variable("mdf")
        x = y + lnC + [B]
        lp = Model(name="MDF_PRIMAL")

        cnstr_names = (
            [f"driving_force_{j:02d}" for j in range(self.Nr_active)]
            + [f"covariance_var_ub_{j:02d}" for j in range(self.Nr)]
            + [f"covariance_var_lb_{j:02d}" for j in range(self.Nr)]
            + [f"log_conc_ub_{j:02d}" for j in range(self.Nc)]
            + [f"log_conc_lb_{j:02d}" for j in range(self.Nc)]
        )

        constraints = []
        for j in range(A.shape[0]):
            row = [A[j, i] * x[i] for i in range(A.shape[1])]
            constraints.append(
                Constraint(sum(row), ub=b[j], name=cnstr_names[j])
            )

        lp.add(constraints)

        row = [c[i] * x[i] for i in range(c.shape[0])]
        lp.objective = Objective(sum(row), direction="max")

        return lp, y, lnC, B

    def _MakeMDFProblemDual(
        self,
    ) -> Tuple[
        Model, List[Variable], List[Variable], List[Variable], List[Variable]
    ]:
        """Create dual LP problem for Min-max Thermodynamic Driving Force.

        Does not set the objective function... leaves that to the caller.

        :return: the linear problem object, and the four types of variables
        as arrays.
        """
        A, b, c, w, g, z, u = self._GetDualVariablesAndConstants()
        x = w + g + z + u
        lp = Model(name="MDF_DUAL")

        cnstr_names = (
            ["y_%02d" % j for j in range(self.Nr)]
            + ["l_%02d" % j for j in range(self.Nc)]
            + ["MDF"]
        )

        constraints = []
        for i in range(A.shape[1]):
            row = [A[j, i] * x[j] for j in range(A.shape[0])]
            constraints.append(
                Constraint(sum(row), lb=c[i], ub=c[i], name=cnstr_names[i])
            )

        lp.add(constraints)

        row = [b[i] * x[i] for i in range(A.shape[0])]
        lp.objective = Objective(sum(row), direction="min")

        return lp, w, g, z, u

    def FindMDF(self) -> PathwayMDFData:
        """Find the MDF.

        :return: a PathwayMDFData object with the results of MDF analysis.
        """

        def get_primal_array(l):
            return np.array([v.primal for v in l], ndmin=1)

        lp_primal, y, lnC, B = self._MakeMDFProblem()

        if lp_primal.optimize() != "optimal":
            logging.warning("LP status %s", lp_primal.status)
            raise Exception("Cannot solve MDF primal optimization problem")

        y = get_primal_array(y)  # covariance eigenvalue prefactors
        lnC = get_primal_array(lnC)  # log concentrations
        B = lp_primal.variables["mdf"].primal

        lp_dual, w, g, z, u = self._MakeMDFProblemDual()
        if lp_dual.optimize() != "optimal":
            raise Exception("cannot solve MDF dual")

        primal_obj = lp_primal.objective.value
        dual_obj = lp_dual.objective.value
        if abs(primal_obj - dual_obj) > 1e-3:
            raise Exception(
                f"Primal != Dual ({primal_obj:.5f} != " f"{dual_obj:.5f}"
            )

        w = get_primal_array(w)
        z = get_primal_array(z)
        u = get_primal_array(u)
        reaction_prices = w
        compound_prices = z - u

        return PathwayMDFData(
            self, B, self.I_dir, lnC, y, reaction_prices, compound_prices
        )



import collections
import typing

import numpy as np
from equilibrator_api import Q_, R, default_T
from scipy import stats


RT = R * default_T

CELL_VOL_PER_DW = Q_(
    "2.7 milliliter / gram"
)  # cell dry weight, [Winkler and Wilson, 1966, http://www.jbc.org/content/241/10/2200.full.pdf+html]

ECF_DEFAULTS = {
    "flux_unit": "mM/s",
    "version": "3",  # options are: 1, 2, 3, or 4
    "kcat_source": "gmean",  # options are: 'fwd' or 'gmean'
    "denominator": "CM",  # options are: 'S', 'SP', '1S', '1SP', or 'CM'
    "regularization": "volume",  # options are: None, 'volume', or 'quadratic'
    "standard_concentration": "1 M",  # must be explicit in order to avoid
    # confusion with other scripts that
    # use different conventions (e.g. 1 mM)
}
str2bool = lambda x: x not in [0, "false", "False"]


def CastToColumnVector(v):
    """
        casts any numeric list of floats to a 2D-matrix with 1 column,
        and rows corresponding to the length of the list

        if the input is a NumPy array or matrix, it will be reshaped to
    """
    if type(v) in [np.ndarray, np.matrix]:
        return np.array(
            np.reshape(v, (np.prod(v.shape), 1)), dtype=float, ndmin=2
        )
    if isinstance(v, collections.Iterable):
        return np.array(list(v), dtype=float, ndmin=2).T
    else:
        raise ValueError(
            "Can only cast lists or numpy arrays, not " + str(type(v))
        )


def QuantitiesToColumnVector(l: typing.Iterable[Q_], unit: str = ""):
    """
        Returns a float column vector, after converting all values to
        the provided unit.
    """
    return np.array(
        [x.m_as(unit) if not np.isnan(x) else np.nan for x in l],
        dtype=float,
        ndmin=2,
    ).T


def PlotCorrelation(ax, x, y, labels, mask=None, scale="log", grid=True):
    """
        scale - if 'log' indicates that the regression should be done on the
                logscale data.
    """
    x = CastToColumnVector(x)
    y = CastToColumnVector(y)

    if mask is None:
        mask = (np.nan_to_num(x) > 0) & (np.nan_to_num(y) > 0)

    ax.grid(grid)
    if scale == "log":
        ax.set_xscale("log")
        ax.set_yscale("log")
        logx = np.log10(x[mask])
        logy = np.log10(y[mask])
        slope, intercept, r_value, p_value, std_err = stats.linregress(
            logx.flat, logy.flat
        )
        rmse = np.sqrt(np.power(logx - logy, 2).mean())
    else:
        ax.set_xscale("linear")
        ax.set_yscale("linear")
        slope, intercept, r_value, p_value, std_err = stats.linregress(
            x[mask].flat, y[mask].flat
        )
        rmse = np.sqrt(np.power(x[mask] - y[mask], 2).mean())
    ax.plot(x[mask], y[mask], ".", markersize=15, color="red", alpha=0.5)
    ax.plot(x[~mask], y[~mask], ".", markersize=15, color="blue", alpha=0.5)

    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    ax.set_xlim(min(xmin, ymin), max(xmax, ymax))
    ax.set_ylim(min(xmin, ymin), max(xmax, ymax))
    ax.plot(
        [0, 1], [0, 1], ":", color="black", alpha=0.4, transform=ax.transAxes
    )

    boxtext = "RMSE = %.2f\n$r^2$ = %.2f\n(p = %.1e)" % (
        rmse,
        r_value ** 2,
        p_value,
    )

    ax.text(
        0.05,
        0.95,
        boxtext,
        verticalalignment="top",
        horizontalalignment="left",
        transform=ax.transAxes,
        color="black",
        fontsize=10,
        bbox={"facecolor": "white", "alpha": 0.5, "pad": 10},
    )

    for l, x_i, y_i, m in zip(labels, x, y, mask):
        if m:
            ax.text(x_i, y_i, l, alpha=1.0)
        elif np.isfinite(x_i) and np.isfinite(y_i):
            if scale == "linear" or (x_i > 0 and y_i > 0):
                ax.text(x_i, y_i, l, alpha=0.4)



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import codecs
import datetime
import os
import types
import xml.dom.minidom

import numpy as np


def _mkdir(newdir):
    """works the way a good mkdir should :)
        - already exists, silently complete
        - regular file in the way, raise an exception
        - parent directory(ies) does not exist, make them as well
    """
    if os.path.isdir(newdir):
        pass
    elif os.path.isfile(newdir):
        raise OSError(
            "a file with the same name as the desired "
            "dir, '%s', already exists." % newdir
        )
    else:
        head, tail = os.path.split(newdir)
        if head and not os.path.isdir(head):
            _mkdir(head)
        if tail:
            os.mkdir(newdir)


class BaseHtmlWriter:
    def __init__(self):
        self.div_counter = 0
        pass

    def relative_to_full_path(self, relpath):
        raise Exception("class not implemented")

    def write(self, s):
        raise Exception("class not implemented")

    def write_header(self):
        self.write(
            '<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">\n'
        )
        self.write("<head>\n")
        self.write(
            '<script type="text/javascript" src="expandCollapse.js"></script>\n'
        )
        self.write("</head>\n")
        self.write("<html>\n<body>\n")

        now = datetime.datetime.now()
        self.write("<div>Written at %s</div>" % now)

    def write_js(self, path):
        if os.path.exists(path + "/expandCollapse.js"):
            return
        fp = open(path + "/expandCollapse.js", "w")
        fp.write(
            """function toggleMe(a){
  var e=document.getElementById(a);
  if(!e)return true;
  if(e.style.display=="none"){
    e.style.display="block"
  } else {
    e.style.display="none"
  }
  return true;
}
"""
        )
        fp.close()

    def write_ol(self, l):
        self.write("<ol>\n")
        for mem in l:
            self.write("  <li>%s</li>\n" % str(mem))
        self.write("</ol>\n")

    def write_ul(self, l):
        self.write("<ul>\n")
        for mem in l:
            self.write("  <li>%s</li>\n" % str(mem))
        self.write("</ul>\n")

    def write_table(self, rowdicts, headers=None, border=1, decimal=None):
        """
            In order to print the row number, use the title '#' in headers and
            write_table() will automatically fill that column with the row numbers.
        """

        def to_string(x, decimal=None):
            if type(x) == types.StringType:
                return x

            if type(x) in (types.IntType, np.int16, np.int32, np.int64):
                return "%d" % x

            if type(x) in (types.FloatType, np.float32, np.float64):
                if np.isnan(x):
                    return "N/A"
                if decimal is not None:
                    return eval("'%%.%df' %% x" % decimal)
                return "%g" % x

            return str(x)

        if not headers:
            headers = set()
            for rowdict in rowdicts:
                for key in rowdict.keys():
                    headers.add(to_string(key))
            headers = sorted(headers)

        self.write("<table border=%d>\n" % border)
        self.write(
            "<tr><td><b>"
            + "</b></td><td><b>".join(headers)
            + "</b></td></tr>\n"
        )
        for i, rowdict in enumerate(rowdicts):
            rowdict["#"] = "%d" % i
            values = [
                to_string(rowdict.get(key, ""), decimal) for key in headers
            ]
            self.write("<tr><td>" + "</td><td>".join(values) + "</td></tr>\n")
        self.write("</table>\n")

    def table_start(self, border=1):
        self.write("<table border=%d>\n" % border)

    def table_writerow(self, values):
        self.write("<tr><td>" + "</td><td>".join(values) + "</td></tr>\n")

    def table_end(self):
        self.write("</table>\n")

    def insert_toggle(self, div_id=None, start_here=False, label="Show"):
        if not div_id:
            div_id = "DIV%05d" % self.div_counter
            self.div_counter += 1
        elif type(div_id) != types.StringType:
            raise ValueError("HTML div ID must be a string")
        self.write(
            '<input type="button" class="button" onclick="return toggleMe(\'%s\')" value="%s">\n'
            % (div_id, label)
        )
        if start_here:
            self.div_start(div_id)
        return div_id

    def div_start(self, div_id):
        self.write('<div id="%s" style="display:none">' % div_id)

    def div_end(self):
        self.write("</div>\n")

    def embed_img(self, fig_fname, alternative_string=""):
        self.write(
            '<img src="' + fig_fname + '" atl="' + alternative_string + '" />'
        )

    def embed_svg(self, fig_fname, width=320, height=240, name=""):
        self.write('<a href="%s.svg">' % name)
        self.extract_svg_from_file(fig_fname, width=width, height=height)
        self.write("</a>")

        # self.write('<object data="%s" type="image/svg+xml" width="%dpt" height="%dpt" name="%s" frameborder="0" marginwidth="0" marginheight="0"/></object>'
        #           % (fig_fname, width, height, name))

    def embed_matplotlib_figure(self, fig, width=None, height=None, name=None):
        """
            Adds a matplotlib figure into the HTML as an inline SVG
            
            Arguments:
                fig          - a matplotlib Figure object
                width        - the desired width of the figure in pixels
                height       - the desired height of the figure in pixels
                name         - if not None, the SVG will be written to a file with that name will
                               be linked to from the inline figure
        """
        if name:
            svg_filename = self.relative_to_full_path(name + ".svg")
            self.write('<a href="%s.svg">' % name)
        else:
            svg_filename = ".svg"

        width = width or (fig.get_figwidth() * fig.get_dpi())
        height = height or (fig.get_figheight() * fig.get_dpi())

        fig.savefig(svg_filename, format="svg")
        self.extract_svg_from_file(svg_filename, width=width, height=height)

        if name:
            self.write("</a>")
        else:
            os.remove(svg_filename)

    def embed_dot_inline(self, Gdot, width=320, height=240, name=None):
        """
            Converts the DOT graph to an SVG DOM and uses the inline SVG option to 
            add it directly into the HTML (without creating a separate SVG file).
        """
        if name:
            svg_filename = self.relative_to_full_path(name + ".svg")
            self.write('<a href="%s.svg">' % name)
        else:
            svg_filename = ".svg"

        Gdot.write(svg_filename, prog="dot", format="svg")
        self.extract_svg_from_file(svg_filename, width=width, height=height)
        if name:
            self.write("</a>")
        else:
            os.remove(svg_filename)

    def embed_dot(self, Gdot, name, width=320, height=240):
        """
            Converts the DOT graph to an SVG DOM and uses the inline SVG option to 
            add it directly into the HTML (without creating a separate SVG file).
        """
        svg_filename = self.relative_to_full_path(name + ".svg")
        Gdot.write(svg_filename, prog="dot", format="svg")
        self.embed_svg(svg_filename, width=width, height=height, name=name)

    def extract_svg_from_xmldom(self, dom, width=320, height=240):
        svg = dom.getElementsByTagName("svg")[0]
        svg.setAttribute("width", "%dpt" % width)
        svg.setAttribute("height", "%dpt" % height)
        self.write(svg.toxml())

    def extract_svg_from_file(self, fname, width=320, height=240):
        xmldom = xml.dom.minidom.parse(fname)
        self.extract_svg_from_xmldom(xmldom, width, height)

    def branch(self, relative_path, link_text=None):
        """
            Branches the HTML file by creating a new HTML and adding a link to it with the desired text
        """
        if link_text == None:
            link_text = relative_path

        self.write('<a href="' + relative_path + '.html">' + link_text + "</a>")
        return HtmlWriter(os.path.join(self.filepath, relative_path + ".html"))

    def close(self):
        self.write("</body>\n</html>\n")


class NullHtmlWriter(BaseHtmlWriter):
    def __init__(self):
        BaseHtmlWriter.__init__(self)
        self.filename = None

    def write(self, s):
        pass

    def relative_to_full_path(self, relpath):
        pass


class HtmlWriter(BaseHtmlWriter):
    def __init__(self, filename, force_path_creation=True, flush_always=True):
        BaseHtmlWriter.__init__(self)
        self.filename = filename
        self.filepath = os.path.dirname(filename)
        self.flush_always = flush_always
        if not os.path.exists(self.filepath):
            if force_path_creation and not os.path.exists(self.filepath):
                _mkdir(self.filepath)
            else:
                raise Exception(
                    "cannot write to HTML file %s since the directory doesn't exist"
                    % filename
                )

        self.file = codecs.open(self.filename, "w", encoding="utf_8")
        self.write_header()
        self.write_js(self.filepath)

    def relative_to_full_path(self, relpath):
        return self.filepath + "/" + relpath

    def write(self, s):
        if self.file == None:
            raise Exception(
                "cannot write to this HTML since it is already closed"
            )
        self.file.write(s)
        if self.flush_always:
            self.file.flush()

    def __del__(self):
        if self.file:
            HtmlWriter.close(self)

    def close(self):
        BaseHtmlWriter.close(self)
        self.file.flush()
        self.file.close()
        self.file = None



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from typing import Tuple

import numpy as np
from equilibrator_api import Q_
from optlang.glpk_interface import Constraint, Model, Objective, Variable
from scipy.optimize import minimize

from .util import ECF_DEFAULTS, RT


QUAD_REGULARIZATION_COEFF = 0.2
METABOLITE_WEIGHT_CORRECTION_FACTOR = 1


class EnzymeCostFunction(object):

    ECF_LEVEL_NAMES = [
        "capacity [M]",
        "thermodynamic",
        "saturation",
        "allosteric",
    ]

    def __init__(
        self,
        S: np.ndarray,
        fluxes: np.ndarray,
        kcat: np.ndarray,
        standard_dg: np.ndarray,
        KMM: np.ndarray,
        ln_conc_lb: np.ndarray,
        ln_conc_ub: np.ndarray,
        mw_enz=None,
        mw_met=None,
        A_act=None,
        A_inh=None,
        K_act=None,
        K_inh=None,
        params=None,
    ):
        """
            Construct a toy model with N intermediate metabolites
            (and N+1 reactions)

            Arguments:
                S        - stoichiometric matrix [unitless]
                v        - steady-state fluxes [M/s]
                kcat     - turnover numbers [1/s]
                kcat_source - either 'gmean' or 'fwd'
                dG0      - standard Gibbs free energies of reaction [kJ/mol]
                KMM      - Michaelis-Menten coefficients [M]
                ln_conc_lb - lower bounds on metabolite concentrations [ln M]
                ln_conc_ub - upper bounds on metabolite concentrations [ln M]
                mw_enz   - enzyme molecular weight [Da]
                mw_met   - metabolite molecular weight [Da]
                A_act    - Hill coefficient matrix of
                           allosteric activators [unitless]
                A_inh    - Hill coefficient matrix of
                           allosteric inhibitors [unitless]
                K_act    - affinity coefficient matrix of
                           allosteric activators [M]
                K_inh    - affinity coefficient matrix of
                           allosteric inhibitors [M]
                params   - dictionary of extra parameters
        """

        self.params = dict(ECF_DEFAULTS)
        if params is not None:
            self.params.update(params)

        self.S = S

        if fluxes.check("[concentration]/[time]"):
            self.fluxes = fluxes.m_as("M/s").flatten()
        elif fluxes.unitless:
            # relative fluxes are dimensionless
            self.fluxes = fluxes.m_as("").flatten()
        else:
            raise ValueError("Fluxes must be in units of M/s or dimensionless")

        assert kcat.check("1/[time]")
        self.kcat = kcat.m_as("1/s").flatten()

        assert standard_dg.check("[energy]/[substance]")
        self.standard_dg_over_rt = (standard_dg / RT).m_as("").flatten()

        assert KMM.check("[concentration]")
        self.KMM = KMM.m_as("M")

        self.ln_conc_lb = ln_conc_lb.flatten()
        self.ln_conc_ub = ln_conc_ub.flatten()

        self.Nc, self.Nr = S.shape
        assert self.fluxes.shape == (self.Nr,)
        assert self.kcat.shape == (self.Nr,)
        assert self.standard_dg_over_rt.shape == (self.Nr,)
        assert self.KMM.shape == (self.Nc, self.Nr)
        assert self.ln_conc_lb.shape == (self.Nc,)
        assert self.ln_conc_ub.shape == (self.Nc,)

        self.cids = ["C%04d" % i for i in range(self.Nc)]

        self.S_subs = abs(self.S)
        self.S_prod = abs(self.S)
        self.S_subs[self.S > 0] = 0
        self.S_prod[self.S < 0] = 0

        # if the kcat source is 'gmean' we need to recalculate the
        # kcat_fwd using the formula:
        # kcat_fwd = kcat_gmean * sqrt(kEQ * prod_S(KMM) / prod_P(KMM))

        if self.params["kcat_source"] == "gmean":
            ln_KMM_prod = np.array(np.diag(self.S.T @ np.log(self.KMM)))
            ln_ratio = -ln_KMM_prod - self.standard_dg_over_rt
            factor = np.sqrt(np.exp(ln_ratio))
            self.kcat *= factor

        # molecular weights of enzymes and metabolites
        if mw_enz is None:
            self.mw_enz = np.ones(self.Nr)
        else:
            assert mw_met.check("[mass]")
            self.mw_enz = mw_enz.m_as("Da").flatten()
            assert self.mw_enz.shape == (self.Nr,)

        if mw_met is None:
            self.mw_met = np.ones(self.Nc)
        else:
            assert mw_met.check("[mass]")
            self.mw_met = mw_met.m_as("Da").flatten()
            assert self.mw_met.shape == (self.Nc,)

        # allosteric regulation term

        if A_act is None or K_act is None:
            self.A_act = np.zeros(S.shape)
            self.K_act = np.ones(S.shape)
        else:
            assert S.shape == A_act.shape
            assert S.shape == K_act.shape
            assert A_act.check("[concentration]")
            assert K_act.check("[concentration]")
            self.A_act = A_act.m_as("M")
            self.K_act = K_act.m_as("M")

        if A_inh is None or K_inh is None:
            self.A_inh = np.zeros(S.shape)
            self.K_inh = np.ones(S.shape)
        else:
            assert S.shape == A_inh.shape
            assert S.shape == K_inh.shape
            assert A_inh.check("[concentration]")
            assert K_inh.check("[concentration]")
            self.A_inh = A_inh.m_as("M")
            self.K_inh = K_inh.m_as("M")

        # preprocessing: these auxiliary matrices help calculate the ECF3 and
        # ECF4 faster
        self.D_S_coeff = np.diag(self.S_subs.T @ np.log(self.KMM))
        self.D_P_coeff = np.diag(self.S_prod.T @ np.log(self.KMM))
        self.act_denom = np.diag(self.A_act.T @ np.log(self.K_act))
        self.inh_denom = np.diag(self.A_inh.T @ np.log(self.K_inh))

        try:
            self.ECF = eval("self._ECF%s" % self.params["version"])
        except AttributeError:
            raise ValueError(
                "The enzyme cost function %d is unknown"
                % self.params["version"]
            )

        try:
            self.D = eval("self._D_%s" % self.params["denominator"])
        except AttributeError:
            raise ValueError(
                "The denominator function %s is unknown"
                % self.params["denominator"]
            )

        self.regularization = self.params["regularization"]

    def _DrivingForce(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            calculate the driving force for every reaction in every condition
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        return (
            -np.tile(self.standard_dg_over_rt, (ln_conc.shape[1], 1)).T
            - self.S.T @ ln_conc
        )

    def _EtaThermodynamic(self, ln_conc: np.ndarray) -> np.ndarray:
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        driving_force = self._DrivingForce(ln_conc)

        # replace infeasbile reactions with a positive driving force to avoid
        # negative cost in ECF2
        eta_thermo = 1.0 - np.exp(-driving_force)

        # set the value of eta to a negative number when the reaction is
        # infeasible so it will be easy to find them, and also calculating
        # 1/x will not return an error
        eta_thermo[driving_force <= 0] = -1.0
        return eta_thermo

    def _D_S(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            return a matrix containing the values of D_S
            i.e. prod(s_i / K_i)^n_i

            each row corresponds to a reaction in the model
            each column corresponds to another set of concentrations (assuming
            lnC is a matrix)
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        return np.exp(
            self.S_subs.T @ ln_conc
            - np.tile(self.D_S_coeff, (ln_conc.shape[1], 1)).T
        )

    def _D_SP(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            return a matrix containing the values of D_SP
            i.e. prod(s_i / K_i)^n_i + prod(p_j / K_j)^n_j

            each row corresponds to a reaction in the model
            each column corresponds to another set of concentrations (assuming
            lnC is a matrix)
        """
        assert ln_conc.shape[0] == self.Nc
        return np.exp(
            self.S_subs.T @ ln_conc
            - np.tile(self.D_S_coeff, (ln_conc.shape[1], 1)).T
        ) + np.exp(
            self.S_prod.T @ ln_conc
            - np.tile(self.D_P_coeff, (ln_conc.shape[1]), 1).T
        )

    def _D_1S(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            return a matrix containing the values of D_1S
            i.e. 1 + prod(s_i / K_i)^n_i

            each row corresponds to a reaction in the model
            each column corresponds to another set of concentrations (assuming
            lnC is a matrix)
        """
        return 1.0 + self._D_S(ln_conc)

    def _D_1SP(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            return a matrix containing the values of D_1SP
            i.e. 1 + prod(s_i / K_i)^n_i + prod(p_j / K_j)^n_j

            each row corresponds to a reaction in the model
            each column corresponds to another set of concentrations (assuming
            lnC is a matrix)
        """
        return 1.0 + self._D_SP(ln_conc)

    def _D_CM(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            return a matrix containing the values of D_CM
            i.e. prod(1 + s_i / K_i)^n_i + prod(1 + p_j / K_j)^n_j - 1

            each row corresponds to a reaction in the model
            each column corresponds to another set of concentrations (assuming
            lnC is a matrix)
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        D = np.zeros((self.Nr, ln_conc.shape[1]))
        for k in range(ln_conc.shape[1]):
            X_k = np.log(
                np.exp(np.tile(ln_conc[:, k], (self.Nr, 1)).T) / self.KMM + 1.0
            )
            ln_1_plus_S = np.diag(self.S_subs.T @ X_k)
            ln_1_plus_P = np.diag(self.S_prod.T @ X_k)
            D[:, k] = np.exp(ln_1_plus_S) + np.exp(ln_1_plus_P) - 1.0
        return D

    def _EtaKinetic(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            the kinetic part of ECF3 and ECF4
        """
        return self._D_S(ln_conc) / self.D(ln_conc)

    def _EtaAllosteric(self, ln_conc: np.ndarray) -> np.ndarray:
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        kin_act = np.exp(
            -self.A_act.T @ ln_conc
            + np.tile(self.act_denom, (ln_conc.shape[1], 1)).T
        )
        kin_inh = np.exp(
            self.A_inh.T @ ln_conc
            - np.tile(self.inh_denom, (ln_conc.shape[1], 1)).T
        )
        eta_kin = 1.0 / (1.0 + kin_act) / (1.0 + kin_inh)
        return eta_kin

    def IsFeasible(self, ln_conc: np.ndarray) -> bool:
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        df = self._DrivingForce(ln_conc)
        return (df > 0).all()

    def GetVmax(self, E):
        """
            calculate the maximal rate of each reaction, kcat is in
            umol/min/mg and E is in gr, so we multiply by 1000

            Returns:
                Vmax  - in units of [umol/min]
        """
        assert E.shape == (self.Nr,)
        return self.kcat * E  # in M/s

    def _ECF1(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                The most basic Enzyme Cost Function (only dependent on flux
                and kcat). Gives the predicted enzyme concentrations in [M]
        """
        # lnC is not used for ECF1, except to determine the size of the result
        # matrix.
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        ecf1 = np.tile(self.fluxes / self.kcat, (ln_conc.shape[1], 1)).T

        return ecf1

    def _ECF2(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                The thermodynamic-only Enzyme Cost Function.
                Gives the predicted enzyme concentrations in [M].
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        ecf2 = self._ECF1(ln_conc) / self._EtaThermodynamic(ln_conc)

        # fix the "fake" values that were given in ECF2 to infeasible reactions
        ecf2[ecf2 < 0] = np.nan

        return ecf2

    def _ECF3(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                An Enzyme Cost Function that integrates kinetic and
                thermodynamic data, but no allosteric regulation.
                Gives the predicted enzyme concentrations in [M].
        """
        # calculate the product of all substrates and products for the kinetic
        # term
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        ecf3 = self._ECF2(ln_conc) / self._EtaKinetic(ln_conc)

        return ecf3

    def _ECF4(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                The full Enzyme Cost Function, i.e. with kinetic, thermodynamic
                and allosteric data.
                Gives the predicted enzyme concentrations in [M].
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        ecf4 = self._ECF3(ln_conc) / self._EtaAllosteric(ln_conc)

        return ecf4

    def GetEnzymeCostPartitions(self, ln_conc: np.ndarray) -> np.ndarray:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                A matrix contining the enzyme costs separated to the 4 ECF
                factors (as columns).
                The first column is the ECF1 predicted concentrations in [M].
                The other columns are unitless (added cost, always > 1)
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        cap = self._ECF1(ln_conc)  # capacity
        trm = 1.0 / self._EtaThermodynamic(ln_conc)  # thermodynamics
        kin = 1.0 / self._EtaKinetic(ln_conc)  # kinetics
        alo = 1.0 / self._EtaAllosteric(ln_conc)  # allostery
        return np.hstack([cap, trm, kin, alo])

    def GetVolumes(self, ln_conc: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        """
            Arguments:
                A single metabolite ln-concentration vector

            Returns:
                Two arrays containing the enzyme volumes and
                metabolite volumes (at the provided point)
        """
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc

        enz_conc = self.ECF(ln_conc)
        met_conc = np.exp(ln_conc)
        enz_vol = enz_conc * self.mw_enz
        met_vol = met_conc * self.mw_met
        return enz_vol, met_vol

    def GetFluxes(self, ln_conc: np.ndarray, E: np.ndarray) -> np.ndarray:
        assert len(ln_conc.shape) == 2
        assert ln_conc.shape[0] == self.Nc
        assert E.shape == (self.Nr,)

        v = np.tile(self.GetVmax(E), (ln_conc.shape[1], 1)).T
        v *= self._EtaThermodynamic(ln_conc)
        v *= self._EtaKinetic(ln_conc)
        v *= self._EtaAllosteric(ln_conc)
        return v

    def ECM(self, ln_conc0: np.ndarray = None, n_iter: int = 10) -> np.ndarray:
        """
            Use convex optimization to find the y with the minimal total
            enzyme cost per flux, i.e. sum(ECF(lnC))
        """

        def optfun(ln_conc: np.ndarray) -> float:
            """
                regularization function:
                    d      = x - 0.5 * (x_min + x_max)
                    lambda = median(enzyme cost weights)
                    reg    = 0.01 * lambda * 0.5 * (d.T * d)
            """
            ln_conc = np.array(ln_conc.flatten(), ndmin=2).T

            # if some reaction is not feasible, give a large penalty
            # proportional to the negative driving force.
            minimal_df = self._DrivingForce(ln_conc).min()
            if minimal_df <= 0:
                return 1e20 * abs(minimal_df)

            enz_conc = self.ECF(ln_conc)
            met_conc = np.exp(ln_conc)

            e = float(enz_conc.T @ self.mw_enz)
            m = float(met_conc.T @ self.mw_met)
            if np.isnan(e) or e <= 0:
                raise Exception(
                    "ECF returns NaN although all reactions are feasible"
                )

            if (
                self.regularization is None
                or self.regularization.lower() == "none"
            ):
                return e
            elif self.regularization.lower() == "volume":
                return e + METABOLITE_WEIGHT_CORRECTION_FACTOR * m
            elif self.regularization.lower() == "quadratic":
                d = ln_conc - 0.5 * (ln_conc.min() + ln_conc.max())
                return e + QUAD_REGULARIZATION_COEFF * 0.5 * float(d.T * d)
            else:
                raise Exception(
                    "Unknown regularization: " + self.regularization
                )

        if ln_conc0 is None:
            mdf, ln_conc0 = self.MDF()
            if np.isnan(mdf) or mdf < 0.0:
                raise ValueError(
                    "It seems that the problem is thermodynamically"
                    " infeasible, therefore ECM is not applicable."
                )
        assert ln_conc0.shape == (self.Nc, 1)

        bounds = list(zip(self.ln_conc_lb.flat, self.ln_conc_ub.flat))

        min_res = np.inf
        ln_conc_min = ln_conc0
        for i in range(n_iter):
            ln_conc0_rand = ln_conc0 * (
                1.0 + 0.1 * np.random.rand(ln_conc0.shape[0], 1)
            )
            r = minimize(
                optfun, x0=ln_conc0_rand, bounds=bounds, method="SLSQP"
            )

            if not r.success:
                logging.info(
                    f"iteration #{i}: optimization unsuccessful "
                    f"because of {r.message}, "
                    "trying again"
                )
                continue

            res = optfun(r.x)
            if res < min_res:
                if min_res == np.inf:
                    logging.info("iteration #%d: cost = %.5f" % (i, res))
                else:
                    logging.info(
                        "iteration #%d: cost = %.5f, decrease factor = %.3e"
                        % (i, res, 1.0 - res / min_res)
                    )
                min_res = res
                ln_conc_min = np.array(r.x, ndmin=2).T
            else:
                logging.info(
                    "iteration #%d: cost = %.5f, no improvement" % (i, res)
                )

        return ln_conc_min

    def MDF(self) -> Tuple[Q_, np.ndarray]:
        """
            Find an initial point (x0) for the optimization using MDF.
        """
        Nc, Nr = self.S.shape

        # ln-concentration variables
        ln_conc = [Variable("l%d" % i) for i in range(self.Nc)]
        B = Variable("mdf")

        lp = Model(name="MDF_PRIMAL")

        constraints = []
        for j in range(Nr):
            driving_force = [self.S[i, j] * ln_conc[i] for i in range(Nc)]
            cnstr = Constraint(
                sum(driving_force) + B,
                ub=-self.standard_dg_over_rt[j],
                name=f"driving_force_{j:02d}",
            )
            constraints.append(cnstr)

        for i in range(Nc):
            cnstr = Constraint(
                ln_conc[i],
                lb=self.ln_conc_lb[i],
                ub=self.ln_conc_ub[i],
                name=f"ln_conc_{i:02d}",
            )
            constraints.append(cnstr)

        lp.add(constraints)

        lp.objective = Objective(B, direction="max")

        if lp.optimize() != "optimal":
            logging.warning("LP status %s", lp.status)
            raise Exception("Cannot solve MDF primal optimization problem")

        mdf = B.primal
        lnC0 = np.array([_m.primal for _m in ln_conc], ndmin=2).T
        return mdf * RT, lnC0



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import numpy as np
from scipy.integrate import ode

from .errors import ThermodynamicallyInfeasibleError


class EnzymeCostSimulator(object):
    def __init__(self, ecf):
        self.ecf = ecf

    def Simulate(self, lnC0, E, t_max=1000, dt=1, eps=1e-9):
        """
            Find the steady-state solution for the metabolite concentrations
            given the enzyme abundances

            Arguments:
                E    - enzyme abundances [gr]
                y0   - initial concentration of internal metabolites
                       (default: MDF solution)
                eps  - the minimal change under which the simulation will stop

            Returns:
                v    - the steady state flux
                y    - the steady state internal metabolite concentrations
        """
        assert lnC0.shape == (self.ecf.Nc, 1)
        assert E.shape == (self.ecf.Nr,)

        ln_conc_bounds_diff = self.ecf.ln_conc_ub - self.ecf.ln_conc_lb
        idx_fixed = np.where(ln_conc_bounds_diff <= 1e-5)[0].tolist()
        idx_non_fixed = np.where(ln_conc_bounds_diff > 1e-5)[0].tolist()

        def f(t, y):
            # we only care about the time derivatives of the internal
            # metabolites (i.e. the first and last one are assumed to be
            # fixed in time)
            ln_conc = np.log(np.array(y, ndmin=2).T)
            v = self.ecf.GetFluxes(ln_conc, E)
            dy = self.ecf.S @ v
            dy[idx_fixed, :] = 0
            return dy

        if not self.ecf.IsFeasible(lnC0):
            raise ThermodynamicallyInfeasibleError(lnC0)

        v = self.ecf.GetFluxes(lnC0, E)

        T = np.array([0])
        Y = np.exp(lnC0).T
        V = v.T

        r = ode(f)
        r.set_initial_value(Y.T, 0)

        while (
            r.successful()
            and r.t < t_max
            and (
                r.t < 0.05 * t_max
                or (np.abs(self.ecf.S[idx_non_fixed, :] @ v) > eps).any()
            )
        ):
            r.integrate(r.t + dt)
            v = self.ecf.GetFluxes(np.log(r.y), E)

            T = np.hstack([T, r.t])
            Y = np.vstack([Y, r.y.T])
            V = np.vstack([V, v.T])

        if r.t >= t_max:
            v_inf = np.nan
            lnC_inf = np.nan
        else:
            v_inf = V[-1, 0]
            lnC_inf = np.log(Y[-1, :])

        return v_inf, lnC_inf



"""Define lower and upper bounds on compounds."""
# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.


from typing import Dict, Iterable, TextIO, Tuple

import numpy as np
import pandas as pd
import pkg_resources
from equilibrator_api import (
    Q_,
    Compound,
    ccache,
    default_conc_lb,
    default_conc_ub,
    standard_concentration,
    ureg,
)


class BaseBounds(object):
    """A base class for declaring bounds on things."""

    def Copy(self):
        """Return a (deep) copy of self."""
        raise NotImplementedError

    def GetLowerBound(self, compound: Compound):
        """Get the lower bound for this key.

        :param key: a compound
        """
        raise NotImplementedError

    def GetUpperBound(self, compound: Compound):
        """Get the upper bound for this key.

        :param key: a compound
        """
        raise NotImplementedError

    def GetLowerBounds(self, compounds: Iterable[Compound]) -> Iterable[Q_]:
        """Get the bounds for a set of keys in order.

        :param compounds: an iterable of Compounds
        :return:  an iterable of the lower bounds
        """
        return map(self.GetLowerBound, compounds)

    def GetUpperBounds(self, compounds: Iterable[Compound]) -> Iterable[Q_]:
        """Get the bounds for a set of keys in order.

        :param compounds: an iterable of Compounds
        :return:  an iterable of the upper bounds
        """
        return map(self.GetUpperBound, compounds)

    def GetBoundTuple(self, compound: Compound) -> Tuple[Q_, Q_]:
        """Get both upper and lower bounds for this key.

        :param compound: a Compound object
        :return: a 2-tuple (lower bound, upper bound)
        """
        return self.GetLowerBound(compound), self.GetUpperBound(compound)

    def GetBounds(
        self, compounds: Iterable[Compound]
    ) -> Tuple[Iterable[Q_], Iterable[Q_]]:
        """Get the bounds for a set of compounds.

        :param compounds: an iterable of Compounds
        :return: a 2-tuple (lower bounds, upper bounds)
        """
        bounds = map(self.GetBoundTuple, compounds)
        lbs, ubs = zip(*bounds)
        return lbs, ubs

    @staticmethod
    @ureg.check("[concentration]")
    def conc2ln_conc(b: Q_) -> float:
        """Convert a concentration to log-concentration.

        :param b: a concentration
        :return: the log concentration
        """
        return np.log((b / standard_concentration).m_as(""))

    def GetLnBounds(
        self, compounds: Iterable[Compound]
    ) -> Tuple[Iterable[float], Iterable[float]]:
        """Get the log-bounds for a set of compounds.

        :param compounds: an iterable of Compounds
        :return: a 2-tuple (log lower bounds, log upper bounds)
        """
        lbs, ubs = self.GetBounds(compounds)

        return map(self.conc2ln_conc, lbs), map(self.conc2ln_conc, ubs)

    def GetLnLowerBounds(
        self, compounds: Iterable[Compound]
    ) -> Iterable[float]:
        """Get the log lower bounds for a set of compounds.

        :param compounds: an iterable of Compounds
        :return: an iterable of log lower bounds
        """
        lbs = self.GetLowerBounds(compounds)
        return map(self.conc2ln_conc, lbs)

    def GetLnUpperBounds(
        self, compounds: Iterable[Compound]
    ) -> Iterable[float]:
        """Get the log upper bounds for a set of compounds.

        :param compounds: an iterable of Compounds
        :return: an iterable of log upper bounds
        """
        ubs = self.GetUpperBounds(compounds)
        return map(self.conc2ln_conc, ubs)

    @ureg.check(None, None, "[concentration]", "[concentration]")
    def SetBounds(self, compound: Compound, lb: Q_, ub: Q_):
        """Set bounds for a specific key.

        :param key: a string representation of a KEGG compound ID,
        i.e. C00001 for water
        :param lb: the lower bound value
        :param ub: the upper bound value
        """
        assert lb <= ub
        self.lower_bounds[compound] = lb
        self.upper_bounds[compound] = ub


class Bounds(BaseBounds):
    """Contains upper and lower bounds for various keys.

    Allows for defaults.

    """

    @ureg.check(None, None, None, "[concentration]", "[concentration]")
    def __init__(
        self,
        lower_bounds: Dict[Compound, Q_] = None,
        upper_bounds: Dict[Compound, Q_] = None,
        default_lb: Q_ = default_conc_lb,
        default_ub: Q_ = default_conc_ub,
    ) -> object:
        """Initialize the bounds object.

        :param lower_bounds: a dictionary mapping keys to lower bounds
        :param upper_bounds: a dictionary mapping keys to upper bounds
        :param default_lb: default lower bound to if no specific one is
        provided
        :param default_lb: default upper bound to if no specific one is
        provided
        """
        self.lower_bounds = lower_bounds or dict()
        self.upper_bounds = upper_bounds or dict()
        for b in self.lower_bounds.values():
            assert b.check("[concentration]")
        for b in self.upper_bounds.values():
            assert b.check("[concentration]")

        self.default_lb = default_lb
        self.default_ub = default_ub

    @classmethod
    @ureg.check(None, None, "[concentration]", "[concentration]")
    def from_csv(
        cls,
        f: TextIO,
        default_lb: Q_ = default_conc_lb,
        default_ub: Q_ = default_conc_ub,
    ) -> object:
        """Read bounds from a .csv file.

        Assume that all the concentrations are given in units of molar.

        :param f: an open .csv file stream
        :param default_lb: default lower bound to if no specific one is
        provided
        :param default_ub: default upper bound to if no specific one is
        provided
        :return: a Bounds object
        """
        lbs: Dict[Compound, Q_] = dict()
        ubs: Dict[Compound, Q_] = dict()
        bounds_df = pd.read_csv(f)

        for row in bounds_df.itertuples():
            compound = ccache.get_compound(row.compound_id)
            if compound is None:
                raise ValueError(
                    f"Cannot find this compound accession: "
                    f"{row.compound_id}"
                )
            lbs[compound] = row.lb * ureg.molar
            ubs[compound] = row.ub * ureg.molar

        bounds = Bounds(lbs, ubs, default_lb, default_ub)
        bounds._check_bounds()
        return bounds

    def to_data_frame(self) -> pd.DataFrame:
        """Convert the list of bounds to a Pandas DataFrame.

        :return: a Pandas DataFrame of all bounds
        """
        data = [
            (c, self.GetLowerBound(c), self.GetUpperBound(c))
            for c in self.lower_bounds.keys()
        ]
        return pd.DataFrame(data=data, columns=["compound", "lb", "ub"])

    @classmethod
    @ureg.check(None, None, None, "[concentration]", "[concentration]")
    def from_dataframe(
        cls,
        df: pd.DataFrame,
        bounds_unit: str = None,
        default_lb: Q_ = default_conc_lb,
        default_ub: Q_ = default_conc_ub,
    ) -> Tuple[object, Dict[Compound, str]]:
        """Read bounds from a Pandas DataFrame.

        :param df: a pandas.DataFrame with the bounds
        :param bounds_unit: optional string of the units. If None, assume
        the units provided in each cell in the DataFrame.
        :param default_lb: default lower bound to if no specific one is
        provided
        :param default_ub: default upper bound to if no specific one is
        provided
        :return:
        """
        df.set_index("Compound", inplace=True)

        # legacy code, before we always assumed that the compound IDs are
        # from KEGG
        if "Compound:Identifiers:kegg.compound" in df.columns:
            df["Compound:Identifiers"] = (
                "KEGG:" + df["Compound:Identifiers:kegg.compound"]
            )

        # now, by default, we assume the namespaces are provided as part of the
        # reaction formulas
        if "Compound:Identifiers" not in df.columns:
            raise KeyError(
                "Could not find a column of Compound:Identifiers "
                "in the ConcentrationConstraints table"
            )

        df["Compound"] = df["Compound:Identifiers"].apply(ccache.get_compound)

        if pd.isnull(df.Compound).any():
            accessions_not_found = df.loc[
                pd.isnull(df.Compound), "Compound:Identifiers"
            ]
            error_msg = str(accessions_not_found.to_dict())
            raise KeyError(
                f"Some compounds not found in equilibrator-cache: "
                f"{error_msg}"
            )
        name_to_compound = df.Compound.to_dict()

        df.set_index("Compound", inplace=True)

        if bounds_unit is not None:
            lbs = df["Concentration:Min"].apply(
                lambda x: Q_(float(x), bounds_unit)
            )
            ubs = df["Concentration:Max"].apply(
                lambda x: Q_(float(x), bounds_unit)
            )
        else:
            lbs = df["Concentration:Min"].apply(Q_)
            ubs = df["Concentration:Max"].apply(Q_)

        bounds = Bounds(lbs.to_dict(), ubs.to_dict(), default_lb, default_ub)
        bounds._check_bounds()

        return bounds, name_to_compound

    def _check_bounds(self) -> None:
        assert self.default_lb <= self.default_ub, (
            f"default lower bound ({self.default_lb}) is higher than the "
            f"default upper bound ({self.default_ub})"
        )

        for compound in self.upper_bounds:
            lb = self.GetLowerBound(compound)
            ub = self.GetUpperBound(compound)
            assert lb <= ub, (
                f"lower bound ({self.default_lb}) for {compound} is higher "
                f"than the upper bound ({self.default_ub})"
            )

    def Copy(self) -> object:
        """Return a deep copy of self."""
        new_lb = dict(self.lower_bounds.items())
        new_ub = dict(self.upper_bounds.items())
        return Bounds(new_lb, new_ub, self.default_lb, self.default_ub)

    def GetLowerBound(self, compound: Compound) -> Q_:
        """Get the lower bound for this key.

        :param compound: a compound
        :return: the lower bound
        """
        return self.lower_bounds.get(compound, self.default_lb)

    def GetUpperBound(self, compound: Compound):
        """Get the upper bound for this key.

        :param compound: a compound
        :return: the upper bound
        """
        return self.upper_bounds.get(compound, self.default_ub)

    # the default is generated only at the first call of GetDefaultBounds()
    # the reason to do this, is because looking up all the cofactor compounds
    # in the cache takes about 20 seconds
    DEFAULT_BOUNDS = None

    @staticmethod
    def GetDefaultBounds() -> object:
        """Return the default lower and upper bounds for a pre-determined list.

        :return: a Bounds object with the default values
        """
        if Bounds.DEFAULT_BOUNDS is None:
            COFACTORS_FNAME = pkg_resources.resource_filename(
                "equilibrator_api", "data/cofactors.csv"
            )
            with open(COFACTORS_FNAME, "r") as fp:
                Bounds.DEFAULT_BOUNDS = Bounds.from_csv(fp)
        return Bounds.DEFAULT_BOUNDS



# The MIT License (MIT)
#
# Copyright (c) 2013 Weizmann Institute of Science
# Copyright (c) 2018-2020 Institute for Molecular Systems Biology,
# ETH Zurich
# Copyright (c) 2018-2020 Novo Nordisk Foundation Center for Biosustainability,
# Technical University of Denmark
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from typing import List, Tuple

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import sbtab
from equilibrator_api import Q_, ccache
from sbtab import SBtab

from .colors import ColorMap
from .cost_function import EnzymeCostFunction
from .errors import ThermodynamicallyInfeasibleError
from .html_writer import BaseHtmlWriter
from .pathway import Pathway
from .util import ECF_DEFAULTS, RT, PlotCorrelation, QuantitiesToColumnVector


class ECMmodel(object):

    DATAFRAME_NAMES = {
        "Compound",
        "Reaction",
        "ConcentrationConstraint",
        "Parameter",
        "Flux",
    }

    def __init__(self, pathway: Pathway, param_df: pd.DataFrame):
        self.config_dict = dict(ECF_DEFAULTS)
        self.config_dict.update(pathway.config_dict)

        stdev_factor = self.config_dict.get("stdev_factor", 1.0)

        (
            rid2crc_gmean,
            rid2crc_fwd,
            rid2crc_rev,
            rid_cid2KMM,
            rid2mw,
            cid2mw,
        ) = ECMmodel.ReadParameters(param_df, self.config_dict)

        self.reaction_ids = list(pathway.reaction_ids)
        self.compound_ids = list(pathway.compound_names)
        assert pathway.standard_dg_primes.check("kJ/mol")
        standard_dg_primes = np.reshape(
            pathway.standard_dg_primes, (len(self.reaction_ids), 1)
        )

        KMM = ECMmodel._GenerateKMM(
            self.compound_ids, self.reaction_ids, rid_cid2KMM
        )

        # we need all fluxes to be positive, so for every negative flux,
        # we multiply it and the corresponding column in S by (-1)
        dir_mat = np.diag(np.sign(pathway.fluxes.magnitude + 1e-10).flat)
        flux = dir_mat @ pathway.fluxes
        S = pathway.S.values @ dir_mat
        standard_dg = dir_mat @ standard_dg_primes

        # we only need to define get kcat in the direction of the flux
        # if we use the 'gmean' option, that means we assume we only know
        # the geometric mean of the kcat, and we distribute it between
        # kcat_fwd and kcat_bwd according to the Haldane relationship
        # if we use the 'fwd' option, we just take the kcat in the
        # direction of flux (as is) and that would mean that our
        # thermodynamic rate law would be equivalent to calculating the
        # reverse kcat using the Haldane relationship
        if self.config_dict["kcat_source"] == "gmean":
            kcat = map(rid2crc_gmean.get, self.reaction_ids)
        elif self.config_dict["kcat_source"] == "fwd":
            # get the relevant kcat (fwd/rev) depending on the directions
            kcat = []
            for rid, d in zip(pathway.reaction_ids, np.diag(dir_mat).flat):
                if d > 0:
                    kcat.append(rid2crc_fwd[rid])
                else:
                    kcat.append(rid2crc_rev[rid])
        else:
            raise ValueError(
                "unrecognized kcat source: " + self.config_dict["kcat_source"]
            )
        kcat = QuantitiesToColumnVector(kcat, "1/s") * Q_(1, "1/s")

        # TODO: turn this into a warning, if the MW data is missing, try to
        # assign a default value
        mw_enz = []
        for rid in self.reaction_ids:
            if rid not in rid2mw:
                raise KeyError(f"This reaction is missing an enzyme MW: {rid}")
            mw_enz.append(rid2mw[rid])
        mw_enz = QuantitiesToColumnVector(mw_enz, "Da") * Q_("Da")

        # TODO: fill gaps in MW data by using the equilibrator-cache
        mw_met = []
        for cid in self.compound_ids:
            if cid not in cid2mw:
                raise KeyError(f"This compound is missing a MW: {cid}")
            mw_met.append(cid2mw[cid])
        mw_met = QuantitiesToColumnVector(mw_met, "Da") * Q_("Da")

        # we must remove H2O from the model, since it should not be considered
        # as a "normal" metabolite in terms of the enzyme and metabolite costs
        idx_water = pathway.S.index.tolist().index(ccache.water)

        self.compound_ids.pop(idx_water)
        S = np.delete(S, idx_water, axis=0)

        # numpy.delete does not support Quantity ndarrays as input
        KMM = np.vstack([KMM[:idx_water, :], KMM[idx_water + 1 :, :]])

        ln_conc_lb = np.delete(pathway.ln_conc_lb, idx_water, axis=0)
        ln_conc_ub = np.delete(pathway.ln_conc_ub, idx_water, axis=0)
        mw_met = np.vstack([mw_met[:idx_water], mw_met[idx_water + 1 :]])

        self.ecf = EnzymeCostFunction(
            S,
            fluxes=flux,
            kcat=kcat,
            standard_dg=standard_dg,
            KMM=KMM,
            ln_conc_lb=ln_conc_lb,
            ln_conc_ub=ln_conc_ub,
            mw_enz=mw_enz,
            mw_met=mw_met,
            params=self.config_dict,
        )

        self._val_df_dict = None

    @staticmethod
    def from_sbtab(sbtabdoc: sbtab.SBtab.SBtabDocument):
        pathway = Pathway.from_sbtab(sbtabdoc)

        param_sbtab = sbtabdoc.get_sbtab_by_id("Parameter")
        assert param_sbtab, "Missing table 'Parameter' in the SBtab document"
        param_df = param_sbtab.to_data_frame()

        return ECMmodel(pathway, param_df)

    def AddValidationData(
        self, validate_sbtabdoc: sbtab.SBtab.SBtabDocument
    ) -> None:
        conc_sbtab = validate_sbtabdoc.get_sbtab_by_id("Concentration")
        self._met_conc_unit = Q_(conc_sbtab.get_attribute("Unit"))
        assert self._met_conc_unit.check("[concentration]"), (
            f"Metabolite concentration unit is not a [concentration] quantity",
            self._met_conc_unit,
        )

        enzyme_sbtab = validate_sbtabdoc.get_sbtab_by_id("EnzymeConcentration")
        self._enz_conc_unit = Q_(enzyme_sbtab.get_attribute("Unit"))
        assert self._enz_conc_unit.check("[concentration]"), (
            f"Enzyme concentration unit is not a [concentration] quantity",
            self._enz_conc_unit,
        )

        self._val_df_dict = {
            sbtab.table_id: sbtab.to_data_frame()
            for sbtab in validate_sbtabdoc.sbtabs
        }

    @staticmethod
    def ReadParameters(
        parameter_df: pd.DataFrame, ecf_params: dict
    ) -> Tuple[dict, ...]:
        cols = ["QuantityType", "Value", "Compound", "Reaction", "Unit"]

        rid2mw = dict()
        cid2mw = dict()
        rid2crc_gmean = dict()  # catalytic rate constant geomertic mean
        rid2crc_fwd = dict()  # catalytic rate constant forward
        rid2crc_rev = dict()  # catalytic rate constant reverse
        crctype2dict = {
            "catalytic rate constant geometric mean": rid2crc_gmean,
            "substrate catalytic rate constant": rid2crc_fwd,
            "product catalytic rate constant": rid2crc_rev,
        }

        rid_cid2KMM = {}  # Michaelis-Menten constants

        for i, row in parameter_df.iterrows():
            try:
                typ, val, cid, rid, unit = [row[c] for c in cols]
                val = Q_(float(val), unit)

                if typ in crctype2dict:
                    assert val.check(
                        "1/[time]"
                    ), f"rate constants must have inverse time units"
                    val.ito("1/s")
                    crctype2dict[typ][rid] = val
                elif typ == "Michaelis constant":
                    assert val.check(
                        "[concentration]"
                    ), f"Michaelis constants must have concentration units"
                    val.ito("molar")
                    rid_cid2KMM[rid, cid] = val
                elif typ == "protein molecular mass":
                    assert val.check("[mass]"), f"'{typ}' must have mass units"
                    val.ito("Da")
                    rid2mw[rid] = val
                elif typ == "molecular mass":
                    assert val.check("[mass]"), f"'{typ}' must have mass units"
                    val.ito("Da")
                    cid2mw[cid] = val
                else:
                    raise AssertionError(
                        "unrecognized Rate Constant Type: " + typ
                    )
            except AssertionError:
                raise ValueError(
                    "Syntax error in Parameter table, row %d - %s" % (i, row)
                )
        # make sure not to count water as contributing to the volume or
        # cost of a reaction
        return (
            rid2crc_gmean,
            rid2crc_fwd,
            rid2crc_rev,
            rid_cid2KMM,
            rid2mw,
            cid2mw,
        )

    @staticmethod
    def _GenerateKMM(
        cids: List[str], rids: List[str], rid_cid2KMM: dict
    ) -> np.array:
        KMM = np.ones((len(cids), len(rids)))
        for i, cid in enumerate(cids):
            for j, rid in enumerate(rids):
                kmm = rid_cid2KMM.get((rid, cid), Q_(1, "M"))
                KMM[i, j] = kmm.m_as("M")
        return KMM * Q_(1, "M")

    def MDF(self):
        mdf, ln_conc0 = self.ecf.MDF()

        if np.isnan(mdf) or mdf < 0.0:
            logging.error("Negative MDF value: %s" % mdf)
            raise ThermodynamicallyInfeasibleError()
        return ln_conc0

    def ECM(self, ln_conc0=None, n_iter=10):
        if ln_conc0 is None:
            ln_conc0 = self.MDF()
            logging.info("initializing ECM using MDF result")

        return self.ecf.ECM(ln_conc0, n_iter=n_iter)

    def ECF(self, ln_conc):
        return self.ecf.ECF(ln_conc)

    @staticmethod
    def _nanfloat(x):
        if type(x) == float:
            return x
        if type(x) == int:
            return float(x)
        if type(x) == str:
            if x.lower() in ["", "nan"]:
                return np.nan
            else:
                return float(x)
        else:
            raise ValueError("unrecognized type for value: " + str(type(x)))

    @staticmethod
    def _MappingToCanonicalEnergyUnits(unit):
        """
            Assuming the canonical units for concentration are Molar

            Returns:
                A function that converts a single number or string to the
                canonical units
        """
        if unit == "kJ/mol":
            return lambda x: ECMmodel._nanfloat(x)
        if unit == "kcal/mol":
            return lambda x: ECMmodel._nanfloat(x) * 4.184

        raise ValueError("Cannot convert these units to kJ/mol: " + unit)

    def _GetMeasuredMetaboliteConcentrations(self) -> np.ndarray:
        if self._val_df_dict is None:
            raise Exception(
                "cannot validate results because no validation data"
                " was given"
            )

        # assume concentrations are in mM
        met_conc_df = self._val_df_dict["Concentration"].set_index("Compound")
        cid2conc = met_conc_df["Value"].apply(
            lambda x: Q_(float(x if x else 0), self._met_conc_unit)
        )
        return QuantitiesToColumnVector(cid2conc[self.compound_ids], "M")

    def _GetMeasuredEnzymeConcentrations(self):
        if self._val_df_dict is None:
            raise Exception(
                "cannot validate results because no validation data"
                " was given"
            )

        enz_conc_df = self._val_df_dict["EnzymeConcentration"].set_index(
            "Reaction"
        )
        rid2conc = enz_conc_df["Value"].apply(
            lambda x: Q_(float(x if x else 0), self._enz_conc_unit)
        )
        return QuantitiesToColumnVector(rid2conc[self.reaction_ids], "M")

    def _GetVolumeDataForPlotting(self, ln_conc):
        enz_vols, met_vols = self.ecf.GetVolumes(ln_conc)

        enz_data = sorted(zip(enz_vols.flat, self.reaction_ids), reverse=True)
        enz_vols, enz_labels = zip(*enz_data)
        enz_colors = [(0.5, 0.8, 0.3)] * len(enz_vols)

        met_data = zip(met_vols.flat, self.compound_ids)
        # remove H2O from the list and sort by descending volume
        met_data = sorted(filter(lambda x: x[1] != "h2o", met_data))
        met_vols, met_labels = zip(*met_data)
        met_colors = [(0.3, 0.5, 0.8)] * len(met_vols)

        return (
            enz_vols + met_vols,
            enz_labels + met_labels,
            enz_colors + met_colors,
        )

    def PlotVolumes(self, ln_conc: np.array, ax: plt.axes) -> None:
        width = 0.8
        vols, labels, colors = self._GetVolumeDataForPlotting(ln_conc)

        ax.bar(np.arange(len(vols)), vols, width, color=colors)
        ax.set_xticklabels(labels, size="medium", rotation=90)
        ax.set_ylabel("total weight [g/L]")

    def PlotVolumesPie(self, ln_conc: np.array, ax: plt.axes) -> None:
        vols, labels, colors = self._GetVolumeDataForPlotting(ln_conc)
        ax.pie(vols, labels=labels, colors=colors)
        ax.set_title("total weight [g/L]")

    def PlotThermodynamicProfile(self, ln_conc: np.array, ax: plt.axes) -> None:
        """
            Plot a cumulative line plot of the dG' values given the solution
            for the metabolite levels. This was originally designed for showing
            MDF results, but is also a useful tool for ECM.
        """
        driving_forces = self.ecf._DrivingForce(ln_conc)

        dgs = [0] + list((-driving_forces).flat)
        cumulative_dgs = np.cumsum(dgs)

        xticks = np.arange(0, len(cumulative_dgs)) - 0.5
        xticklabels = [""] + self.reaction_ids
        ax.plot(cumulative_dgs)
        ax.set_xticks(xticks)
        ax.set_xticklabels(xticklabels, rotation=45, ha="right")
        ax.set_xlim(0, len(cumulative_dgs) - 1)
        ax.set_xlabel("")
        ax.set_ylabel(r"Cumulative $\Delta_r G'$ (kJ/mol)", family="sans-serif")

    def PlotEnzymeDemandBreakdown(
        self,
        ln_conc: np.array,
        ax: plt.Axes,
        top_level: int = 3,
        plot_measured: bool = False,
    ) -> None:
        """
            A bar plot in log-scale showing the partitioning of cost between
            the levels of kinetic costs:
            1 - capacity
            2 - thermodynamics
            3 - saturation
            4 - allosteric
        """
        assert top_level in range(1, 5)

        costs = np.array(self.ecf.GetEnzymeCostPartitions(ln_conc))

        # give all reactions with zero cost a base value, which we will
        # also set as the bottom ylim, which will simulate a "minus infinity"
        # when we plot it in log-scale
        base = min(filter(None, costs[:, 0])) / 2.0
        idx_zero = costs[:, 0] == 0
        costs[idx_zero, 0] = base
        costs[idx_zero, 1:] = 1.0

        bottoms = np.hstack(
            [np.ones((costs.shape[0], 1)) * base, np.cumprod(costs, 1)]
        )
        steps = np.diff(bottoms)

        labels = EnzymeCostFunction.ECF_LEVEL_NAMES[0:top_level]

        ind = range(costs.shape[0])  # the x locations for the groups
        width = 0.8
        ax.set_yscale("log")

        if plot_measured:
            all_labels = ["measured"] + labels
            meas_conc = self._GetMeasuredEnzymeConcentrations()
            cmap = ColorMap(
                all_labels,
                saturation=0.7,
                value=1.0,
                hues=[30.0 / 255, 170.0 / 255, 200.0 / 255, 5.0 / 255],
            )
            ax.plot(
                ind,
                meas_conc,
                color=cmap["measured"],
                marker="o",
                markersize=5,
                linewidth=0,
                markeredgewidth=0.3,
                markeredgecolor=(0.3, 0.3, 0.3),
            )
        else:
            all_labels = labels
            cmap = ColorMap(
                labels,
                saturation=0.7,
                value=0.8,
                hues=[170.0 / 255, 200.0 / 255, 5.0 / 255],
            )

        for i, label in enumerate(labels):
            ax.bar(
                ind,
                steps[:, i].flat,
                width,
                bottom=bottoms[:, i].flat,
                color=cmap[label],
            )

        ax.set_xticks(ind)
        xticks = self.reaction_ids
        ax.set_xticklabels(xticks, size="medium", rotation=90)
        ax.legend(all_labels, loc="best", framealpha=0.2)
        ax.set_ylabel("enzyme demand [M]")
        ax.set_ylim(bottom=base)

    def ValidateMetaboliteConcentrations(
        self, ln_conc: np.array, ax: plt.Axes, scale: str = "log"
    ) -> None:
        pred_conc = np.exp(ln_conc)

        meas_conc = self._GetMeasuredMetaboliteConcentrations()

        # remove NaNs and zeros
        mask = np.nan_to_num(meas_conc) > 0
        mask &= np.nan_to_num(pred_conc) > 0

        mask = mask.flatten()

        # remove compounds with fixed concentrations
        mask &= (self.ecf.ln_conc_ub - self.ecf.ln_conc_lb) > 1e-9

        PlotCorrelation(
            ax, meas_conc, pred_conc, self.compound_ids, mask, scale=scale
        )
        ax.set_xlabel("measured [M]")
        ax.set_ylabel("predicted [M]")

    def ValidateEnzymeConcentrations(
        self, ln_conc: np.array, ax: plt.Axes, scale: str = "log"
    ) -> None:
        pred_conc = self.ecf.ECF(ln_conc)
        meas_conc = self._GetMeasuredEnzymeConcentrations()

        labels = self.reaction_ids
        PlotCorrelation(ax, meas_conc, pred_conc, labels, scale=scale)

        ax.set_xlabel("measured [M]")
        ax.set_ylabel("predicted [M]")

    def to_sbtab(self, ln_conc: np.array) -> sbtab.SBtab.SBtabDocument:
        met_data = []
        for i, cid in enumerate(self.compound_ids):
            met_data.append(("concentration", cid, np.exp(ln_conc[i, 0])))
        met_df = pd.DataFrame(
            columns=["QuantityType", "Compound", "ecm"], data=met_data
        )

        enz_conc = self.ecf.ECF(ln_conc)
        enz_data = []
        for i, rid in enumerate(self.reaction_ids):
            enz_data.append(("concentration of enzyme", rid, enz_conc[i, 0]))
        enz_df = pd.DataFrame(
            columns=["QuantityType", "Reaction", "ecm"], data=enz_data
        )

        sbtabdoc = SBtab.SBtabDocument("report")
        met_sbtab = SBtab.SBtabTable.from_data_frame(
            met_df,
            table_id="Predicted concentrations",
            table_type="Quantity",
            unit="M",
        )

        enz_sbtab = SBtab.SBtabTable.from_data_frame(
            enz_df,
            table_id="Predicted enzyme levels",
            table_type="Quantity",
            unit="M",
        )

        sbtabdoc.add_sbtab(met_sbtab)
        sbtabdoc.add_sbtab(enz_sbtab)
        return sbtabdoc

    def WriteHtmlTables(self, ln_conc: np.array, html: BaseHtmlWriter) -> None:
        meas_enz2conc = self._GetMeasuredEnzymeConcentrations()
        meas_conc = np.array(
            list(
                map(lambda r: meas_enz2conc.get(r, np.nan), self.reaction_ids)
            ),
            ndmin=2,
        ).T
        data_mat = np.hstack(
            [
                self.ecf.fluxes,
                meas_conc,
                self.ecf.ECF(ln_conc),
                self.ecf._DrivingForce(ln_conc),
                self.ecf.GetEnzymeCostPartitions(ln_conc),
            ]
        )

        data_mat[:, 0] *= 1e3  # convert flux from M/s to mM/s
        data_mat[:, 1] *= 1e6  # convert measured enzyme conc. from M to uM
        data_mat[:, 2] *= 1e6  # convert predicted enzyme conc. from M to uM
        data_mat[:, 4] *= 1e6  # convert capacity term from M to uM

        headers = [
            "reaction",
            "flux [mM/s]",
            "measured enz. conc. [uM]",
            "predicted enz. conc. [uM]",
            "driving force [kJ/mol]",
            "capacity [uM]",
        ] + EnzymeCostFunction.ECF_LEVEL_NAMES[1:]
        values = zip(self.reaction_ids, *data_mat.T.tolist())
        values.append(
            [
                "total",
                "",
                data_mat[:, 1].sum(),
                data_mat[:, 2].sum(),
                "",
                "",
                "",
                "",
                "",
            ]
        )

        rowdicst = [dict(zip(headers, v)) for v in values]
        html.write_table(rowdicst, headers=headers, decimal=3)

        meas_met2conc = self._GetMeasuredMetaboliteConcentrations()
        meas_conc = np.array(
            list(
                map(lambda r: meas_met2conc.get(r, np.nan), self.compound_ids)
            ),
            ndmin=2,
        ).T
        data_mat = np.hstack(
            [
                meas_conc,
                np.exp(ln_conc),
                np.exp(self.ecf.ln_conc_lb),
                np.exp(self.ecf.ln_conc_ub),
            ]
        )

        data_mat *= 1e3  # convert all concentrations from M to mM

        headers = [
            "compound",
            "KEGG ID",
            "measured conc. [mM]",
            "predicted conc. [mM]",
            "lower bound [mM]",
            "upper bound [mM]",
        ]
        values = zip(self.compound_ids, *data_mat.T.tolist())
        rowdicst = [dict(zip(headers, v)) for v in values]
        headers = [
            "compound",
            "measured conc. [mM]",
            "lower bound [mM]",
            "predicted conc. [mM]",
            "upper bound [mM]",
        ]
        html.write_table(rowdicst, headers=headers)



# -*- coding: utf-8 -*-
"""
Created on Tue Apr 14 16:41:01 2015

@author: noore
"""

COLORS = "grbcmyk"

import colorsys

import matplotlib.pyplot as plt
import numpy as np


def RgbToHex(rgb):
    r, g, b = rgb
    r *= 255
    g *= 255
    b *= 255
    return "#%02x%02x%02x" % (int(r), int(g), int(b))


def ColorMap(items, saturation=0.7, value=0.95, hues=None):
    if hues is None:
        n = len(items)
        hues = np.arange(float(n)) / float(n)
    f = lambda h: RgbToHex(colorsys.hsv_to_rgb(h, saturation, value))
    rgbs = map(f, hues)
    return dict(zip(items, rgbs))


def LayeredColorMap(
    items, n_minor, saturation=(0.4, 0.7), value=(0.95, 0.5), hues=None
):
    n = len(saturation)
    colormaps = [
        ColorMap(items, saturation[i], value[i], hues) for i in range(n)
    ]
    colormap = {}
    for item in items:
        colormap[item] = [colormaps[i % n][item] for i in range(n_minor)]
    return colormap


def test2():
    items = ["A", "B", "C", "D", "E"]
    hues = [0.05, 0.22, 0.5, 0.7, 0.9]
    n = 5
    m = 10
    colormap = LayeredColorMap(
        items, n, saturation=(0.5, 0.0), value=(0.9, 1.0), hues=hues
    )
    data = abs(np.random.randn(len(items) * n, m))
    bottom = np.zeros((1, m))

    ind = np.arange(m)  # the x locations for the groups
    width = 0.35  # the width of the bars

    for i, item in enumerate(items):
        for j in range(n):
            plt.bar(
                ind,
                data[n * i + j, :],
                width,
                color=colormap[item][j],
                label=item + "%d" % j,
                bottom=bottom[0, :],
                edgecolor="none",
            )
            bottom += data[n * i + j, :]

    top = max(bottom[0, :]) - bottom
    plt.bar(
        ind,
        top[0, :],
        width,
        color="#AAAAAA",
        label=item + "%d" % j,
        bottom=bottom[0, :],
        edgecolor="none",
    )

    plt.legend()
    plt.show()


def test():
    items = ["A", "B", "C", "D", "E"]
    hues = [0.05, 0.22, 0.5, 0.7, 0.9]
    n = 5
    m = 10
    colormap = ColorMap(items, saturation=0.5, value=0.8, hues=hues)
    data = abs(np.random.randn(len(items) * n, m))
    bottom = np.zeros((1, m))
    item_bottom = np.zeros((1, m))

    ind = np.arange(m)  # the x locations for the groups
    width = 0.35  # the width of the bars

    for i, item in enumerate(items):
        for j in range(n):
            if j % 2 == 0:
                facecolor = colormap[item]
            else:
                facecolor = "#FFFFFF"
            plt.bar(
                ind,
                data[n * i + j, :],
                width,
                color=facecolor,
                edgecolor="none",
                label=item + "%d" % j,
                bottom=bottom[0, :],
            )
            bottom += data[n * i + j, :]

    for i, item in enumerate(items):
        height = sum(data[n * i : n * (i + 1), :], 0)
        bottom = sum(data[0 : n * i, :], 0)

        plt.bar(
            ind,
            height,
            width,
            color="none",
            edgecolor=colormap[item],
            label=None,
            bottom=bottom,
        )

    top = max(sum(data, 0)) - sum(data, 0)
    plt.bar(
        ind,
        top,
        width,
        color="#AAAAAA",
        edgecolor="#AAAAAA",
        label=item + "%d" % j,
        bottom=sum(data, 0),
    )

    plt.legend()
    plt.savefig("/home/eladn/Desktop/bars.eps", format="eps")


if __name__ == "__main__":
    test()
